"""
Salt module to manage Unix mounts and the fstab file
"""


import logging
import os
import re

import salt.utils.args
import salt.utils.data
import salt.utils.files
import salt.utils.mount
import salt.utils.path
import salt.utils.platform
import salt.utils.stringutils
from salt.exceptions import CommandExecutionError, CommandNotFoundError
from salt.utils.odict import OrderedDict

# Set up logger
log = logging.getLogger(__name__)

# Define the module's virtual name
__virtualname__ = "mount"


def __virtual__():
    """
    Only load on POSIX-like systems
    """
    # Disable on Windows, a specific file module exists:
    if salt.utils.platform.is_windows():
        return (False, "The mount module cannot be loaded: not a POSIX-like system.")
    return True


def _list_mounts():
    ret = {}
    if __grains__["os"] in ["MacOS", "Darwin"]:
        mounts = __salt__["cmd.run_stdout"]("mount")
    else:
        mounts = __salt__["cmd.run_stdout"]("mount -l")

    for line in mounts.split("\n"):
        comps = re.sub(r"\s+", " ", line).split()
        if len(comps) >= 3:
            ret[comps[2]] = comps[0]
    return ret


def _active_mountinfo(ret):
    _list = _list_mounts()
    filename = "/proc/self/mountinfo"
    if not os.access(filename, os.R_OK):
        msg = "File not readable {0}"
        raise CommandExecutionError(msg.format(filename))

    if "disk.blkid" not in __context__:
        __context__["disk.blkid"] = __salt__["disk.blkid"]()
    blkid_info = __context__["disk.blkid"]

    with salt.utils.files.fopen(filename) as ifile:
        for line in ifile:
            comps = salt.utils.stringutils.to_unicode(line).split()
            device = comps[2].split(":")
            # each line can have any number of
            # optional parameters, we use the
            # location of the separator field to
            # determine the location of the elements
            # after it.
            _sep = comps.index("-")
            device_name = comps[_sep + 2]
            device_uuid = None
            device_label = None
            if device_name:
                device_uuid = blkid_info.get(device_name, {}).get("UUID")
                device_uuid = device_uuid and device_uuid.lower()
                device_label = blkid_info.get(device_name, {}).get("LABEL")
            ret[comps[4]] = {
                "mountid": comps[0],
                "parentid": comps[1],
                "major": device[0],
                "minor": device[1],
                "root": comps[3],
                "opts": _resolve_user_group_names(comps[5].split(",")),
                "fstype": comps[_sep + 1],
                "device": device_name.replace("\\040", "\\ "),
                "alt_device": _list.get(comps[4], None),
                "superopts": _resolve_user_group_names(comps[_sep + 3].split(",")),
                "device_uuid": device_uuid,
                "device_label": device_label,
            }
    return ret


def _active_mounts(ret):
    """
    List active mounts on Linux systems
    """
    _list = _list_mounts()
    filename = "/proc/self/mounts"
    if not os.access(filename, os.R_OK):
        msg = "File not readable {0}"
        raise CommandExecutionError(msg.format(filename))

    with salt.utils.files.fopen(filename) as ifile:
        for line in ifile:
            comps = salt.utils.stringutils.to_unicode(line).split()
            ret[comps[1]] = {
                "device": comps[0],
                "alt_device": _list.get(comps[1], None),
                "fstype": comps[2],
                "opts": _resolve_user_group_names(comps[3].split(",")),
            }
    return ret


def _active_mounts_aix(ret):
    """
    List active mounts on AIX systems
    """
    for line in __salt__["cmd.run_stdout"]("mount -p").split("\n"):
        comps = re.sub(r"\s+", " ", line).split()
        if comps:
            if comps[0] == "node" or comps[0] == "--------":
                continue
            comps_len = len(comps)
            if line.startswith((" ", "\t")):
                curr_opts = (
                    _resolve_user_group_names(comps[6].split(","))
                    if 7 == comps_len
                    else []
                )
                if curr_opts:
                    ret[comps[1]] = {
                        "device": comps[0],
                        "fstype": comps[2],
                        "opts": curr_opts,
                    }
                else:
                    ret[comps[1]] = {"device": comps[0], "fstype": comps[2]}
            else:
                curr_opts = (
                    _resolve_user_group_names(comps[7].split(","))
                    if 8 == comps_len
                    else []
                )
                if curr_opts:
                    ret[comps[2]] = {
                        "node": comps[0],
                        "device": comps[1],
                        "fstype": comps[3],
                        "opts": curr_opts,
                    }
                else:
                    ret[comps[2]] = {
                        "node": comps[0],
                        "device": comps[1],
                        "fstype": comps[3],
                    }
    return ret


def _active_mounts_freebsd(ret):
    """
    List active mounts on FreeBSD systems
    """
    for line in __salt__["cmd.run_stdout"]("mount -p").split("\n"):
        comps = re.sub(r"\s+", " ", line).split()
        ret[comps[1]] = {
            "device": comps[0],
            "fstype": comps[2],
            "opts": _resolve_user_group_names(comps[3].split(",")),
        }
    return ret


def _active_mounts_solaris(ret):
    """
    List active mounts on Solaris systems
    """
    for line in __salt__["cmd.run_stdout"]("mount -v").split("\n"):
        comps = re.sub(r"\s+", " ", line).split()
        ret[comps[2]] = {
            "device": comps[0],
            "fstype": comps[4],
            "opts": _resolve_user_group_names(comps[5].split("/")),
        }
    return ret


def _active_mounts_openbsd(ret):
    """
    List active mounts on OpenBSD systems
    """
    for line in __salt__["cmd.run_stdout"]("mount -v").split("\n"):
        comps = re.sub(r"\s+", " ", line).split()
        parens = re.findall(r"\((.*?)\)", line, re.DOTALL)
        if len(parens) > 1:
            nod = __salt__["cmd.run_stdout"]("ls -l {}".format(comps[0]))
            nod = " ".join(nod.split()).split(" ")
            ret[comps[3]] = {
                "device": comps[0],
                "fstype": comps[5],
                "opts": _resolve_user_group_names(parens[1].split(", ")),
                "major": str(nod[4].strip(",")),
                "minor": str(nod[5]),
                "device_uuid": parens[0],
            }
        else:
            ret[comps[2]] = {
                "device": comps[0],
                "fstype": comps[4],
                "opts": _resolve_user_group_names(parens[0].split(", ")),
            }
    return ret


def _active_mounts_darwin(ret):
    """
    List active mounts on Mac OS systems
    """
    for line in __salt__["cmd.run_stdout"]("mount").split("\n"):
        comps = re.sub(r"\s+", " ", line).split()
        parens = re.findall(r"\((.*?)\)", line, re.DOTALL)[0].split(", ")
        ret[comps[2]] = {
            "device": comps[0],
            "fstype": parens[0],
            "opts": _resolve_user_group_names(parens[1:]),
        }
    return ret


def _resolve_user_group_names(opts):
    """
    Resolve user and group names in related opts
    """
    name_id_opts = {"uid": "user.info", "gid": "group.info"}
    for ind, opt in enumerate(opts):
        if opt.split("=")[0] in name_id_opts:
            _givenid = opt.split("=")[1]
            _param = opt.split("=")[0]
            _id = _givenid
            if not re.match("[0-9]+$", _givenid):
                _info = __salt__[name_id_opts[_param]](_givenid)
                if _info and _param in _info:
                    _id = _info[_param]
            opts[ind] = _param + "=" + str(_id)
        opts[ind] = opts[ind].replace("\\040", "\\ ")
    return opts


def active(extended=False):
    """
    List the active mounts.

    CLI Example:

    .. code-block:: bash

        salt '*' mount.active
    """
    ret = {}
    if __grains__["os"] == "FreeBSD":
        _active_mounts_freebsd(ret)
    elif "AIX" in __grains__["kernel"]:
        _active_mounts_aix(ret)
    elif __grains__["kernel"] == "SunOS":
        _active_mounts_solaris(ret)
    elif __grains__["os"] == "OpenBSD":
        _active_mounts_openbsd(ret)
    elif __grains__["os"] in ["MacOS", "Darwin"]:
        _active_mounts_darwin(ret)
    else:
        if extended:
            try:
                _active_mountinfo(ret)
            except CommandExecutionError:
                _active_mounts(ret)
        else:
            _active_mounts(ret)
    return ret


class _fstab_entry:
    """
    Utility class for manipulating fstab entries. Primarily we're parsing,
    formatting, and comparing lines. Parsing emits dicts expected from
    fstab() or raises a ValueError.

    Note: We'll probably want to use os.normpath and os.normcase on 'name'
    """

    class ParseError(ValueError):
        """
        Error raised when a line isn't parsible as an fstab entry
        """

    fstab_keys = ("device", "name", "fstype", "opts", "dump", "pass_num")

    # preserve data format
    compatibility_keys = ("device", "name", "fstype", "opts", "dump", "pass")

    fstab_format = "{device}\t\t{name}\t{fstype}\t{opts}\t{dump} {pass_num}\n"

    @classmethod
    def dict_from_line(cls, line, keys=fstab_keys):
        if len(keys) != 6:
            raise ValueError("Invalid key array: {}".format(keys))
        if line.startswith("#"):
            raise cls.ParseError("Comment!")

        comps = line.split()
        if len(comps) < 4 or len(comps) > 6:
            raise cls.ParseError("Invalid Entry!")

        comps.extend(["0"] * (len(keys) - len(comps)))

        return dict(zip(keys, comps))

    @classmethod
    def from_line(cls, *args, **kwargs):
        return cls(**cls.dict_from_line(*args, **kwargs))

    @classmethod
    def dict_to_line(cls, entry):
        return cls.fstab_format.format(**entry)

    def __str__(self):
        """
        String value, only works for full repr
        """
        return self.dict_to_line(self.criteria)

    def __repr__(self):
        """
        Always works
        """
        return repr(self.criteria)

    def pick(self, keys):
        """
        Returns an instance with just those keys
        """
        subset = {key: self.criteria[key] for key in keys}
        return self.__class__(**subset)

    def __init__(self, **criteria):
        """
        Store non-empty, non-null values to use as filter
        """
        items = [
            key_value for key_value in criteria.items() if key_value[1] is not None
        ]
        items = [(key_value1[0], str(key_value1[1])) for key_value1 in items]
        self.criteria = dict(items)

    @staticmethod
    def norm_path(path):
        """
        Resolve equivalent paths equivalently
        """
        return os.path.normcase(os.path.normpath(path))

    def match(self, line):
        """
        Compare potentially partial criteria against line
        """
        entry = self.dict_from_line(line)
        for key, value in self.criteria.items():
            if entry[key] != value:
                return False
        return True


class _vfstab_entry:
    """
    Utility class for manipulating vfstab entries. Primarily we're parsing,
    formatting, and comparing lines. Parsing emits dicts expected from
    fstab() or raises a ValueError.

    Note: We'll probably want to use os.normpath and os.normcase on 'name'
    Note: This parses vfstab entries on Solaris like systems

    #device     device      mount       FS  fsck    mount   mount
    #to mount   to fsck     point       type    pass    at boot options
    #
    /devices    -   /devices            devfs    -   no     -
    """

    class ParseError(ValueError):
        """Error raised when a line isn't parsible as an fstab entry"""

    vfstab_keys = (
        "device",
        "device_fsck",
        "name",
        "fstype",
        "pass_fsck",
        "mount_at_boot",
        "opts",
    )
    # NOTE: weird formatting to match default spacing on Solaris
    vfstab_format = (
        "{device:<11} {device_fsck:<3} {name:<19} {fstype:<8} {pass_fsck:<3}"
        " {mount_at_boot:<6} {opts}\n"
    )

    @classmethod
    def dict_from_line(cls, line):
        if line.startswith("#"):
            raise cls.ParseError("Comment!")

        comps = line.split()
        if len(comps) != 7:
            raise cls.ParseError("Invalid Entry!")

        return dict(zip(cls.vfstab_keys, comps))

    @classmethod
    def from_line(cls, *args, **kwargs):
        return cls(**cls.dict_from_line(*args, **kwargs))

    @classmethod
    def dict_to_line(cls, entry):
        return cls.vfstab_format.format(**entry)

    def __str__(self):
        """
        String value, only works for full repr
        """
        return self.dict_to_line(self.criteria)

    def __repr__(self):
        """
        Always works
        """
        return repr(self.criteria)

    def pick(self, keys):
        """
        Returns an instance with just those keys
        """
        subset = {key: self.criteria[key] for key in keys}
        return self.__class__(**subset)

    def __init__(self, **criteria):
        """
        Store non-empty, non-null values to use as filter
        """
        items = [
            key_value for key_value in criteria.items() if key_value[1] is not None
        ]
        items = [(key_value1[0], str(key_value1[1])) for key_value1 in items]
        self.criteria = dict(items)

    @staticmethod
    def norm_path(path):
        """
        Resolve equivalent paths equivalently
        """
        return os.path.normcase(os.path.normpath(path))

    def match(self, line):
        """
        Compare potentially partial criteria against line
        """
        entry = self.dict_from_line(line)
        for key, value in self.criteria.items():
            if entry[key] != value:
                return False
        return True


class _FileSystemsEntry:
    """
    Utility class for manipulating filesystem entries. Primarily we're parsing,
    formatting, and comparing lines. Parsing emits dicts expected from
    fstab() or raises a ValueError.

    Note: We'll probably want to use os.normpath and os.normcase on 'name'
    """

    class ParseError(ValueError):
        """
        Error raised when a line isn't parsible as an fstab entry
        """

    filesystems_keys = ("device", "name", "fstype", "vfstype", "opts", "mount")

    # preserve data format of filesystems
    compatibility_keys = (
        "dev",
        "dev",
        "name",
        "fstype",
        "vfstype",
        "opts",
        "mount",
        "type",
        "vfs",
        "account",
        "boot",
        "check",
        "free",
        "nodename",
        "quota",
        "size",
        "vol",
        "log",
    )

    @classmethod
    def dict_from_lines(cls, lines, keys=filesystems_keys):
        if len(lines) < 2:
            raise ValueError("Invalid number of lines: {}".format(lines))
        if not keys:
            # if empty force default filesystems_keys
            keys = _FileSystemsEntry.filesystems_keys
        elif len(keys) < 6:
            raise ValueError("Invalid key name array: {}".format(keys))

        blk_lines = lines
        orddict = OrderedDict()
        orddict["name"] = blk_lines[0].split(":")[0].strip()
        blk_lines.pop(0)
        for line in blk_lines:
            if line.startswith("#"):
                raise cls.ParseError("Comment!")

            comps = line.split("= ")
            if len(comps) != 2:
                raise cls.ParseError("Invalid Entry!")

            key_name = comps[0].strip()
            if key_name in keys:
                orddict[key_name] = comps[1].strip()
            else:
                raise ValueError(
                    "Invalid name for use in filesystems: {}".format(key_name)
                )

        return orddict

    @classmethod
    def dict_from_cmd_line(cls, ipargs, keys):
        cmdln_dict = ipargs
        if keys:
            for key, value in keys:
                # ignore unknown or local scope keys
                if key.startswith("__"):
                    continue
                if key in _FileSystemsEntry.compatibility_keys:
                    cmdln_dict[key] = value

        return cmdln_dict

    @classmethod
    def from_line(cls, *args, **kwargs):
        return cls(**cls.dict_from_cmd_line(*args, **kwargs))

    @classmethod
    def dict_to_lines(cls, fsys_dict_entry):
        entry = fsys_dict_entry
        strg_out = entry["name"] + ":" + os.linesep
        for k, v in entry.items():
            if "name" not in k:
                strg_out += "\t{}\t\t= {}".format(k, v) + os.linesep
        strg_out += os.linesep
        return str(strg_out)

    def dict_from_entry(self):
        ret = OrderedDict()
        ret[self.criteria["name"]] = self.criteria
        return ret

    def __str__(self):
        """
        String value, only works for full repr
        """
        return self.dict_to_lines(self.criteria)

    def __repr__(self):
        """
        Always works
        """
        return repr(self.criteria)

    def pick(self, keys):
        """
        Returns an instance with just those keys
        """
        subset = {key: self.criteria[key] for key in keys}
        return self.__class__(**subset)

    def __init__(self, **criteria):
        """
        Store non-empty, non-null values to use as filter
        """
        items = [
            key_value for key_value in criteria.items() if key_value[1] is not None
        ]
        items = [(key_value1[0], str(key_value1[1])) for key_value1 in items]
        self.criteria = OrderedDict(items)

    @staticmethod
    def norm_path(path):
        """
        Resolve equivalent paths equivalently
        """
        return os.path.normcase(os.path.normpath(path))

    def match(self, fsys_view):
        """
        Compare potentially partial criteria against built filesystems entry dictionary
        """
        evalue_dict = fsys_view[1]
        for key, value in self.criteria.items():
            if key in evalue_dict:
                if evalue_dict[key] != value:
                    return False
            else:
                return False
        return True

    def __getitem__(self, key):
        """
        Return value for input key
        """
        return self.criteria[key]


def fstab(config="/etc/fstab"):
    """
    .. versionchanged:: 2016.3.2

    List the contents of the fstab

    CLI Example:

    .. code-block:: bash

        salt '*' mount.fstab
    """
    ret = {}
    if not os.path.isfile(config):
        return ret
    with salt.utils.files.fopen(config) as ifile:
        for line in ifile:
            line = salt.utils.stringutils.to_unicode(line)
            try:
                if __grains__["kernel"] == "SunOS":
                    # Note: comments use in default vfstab file!
                    if line[0] == "#":
                        continue
                    entry = _vfstab_entry.dict_from_line(line)
                else:
                    entry = _fstab_entry.dict_from_line(
                        line, _fstab_entry.compatibility_keys
                    )

                entry["opts"] = entry["opts"].split(",")
                while entry["name"] in ret:
                    entry["name"] += "_"

                ret[entry.pop("name")] = entry
            except _fstab_entry.ParseError:
                pass
            except _vfstab_entry.ParseError:
                pass

    return ret


def vfstab(config="/etc/vfstab"):
    """
    .. versionadded:: 2016.3.2

    List the contents of the vfstab

    CLI Example:

    .. code-block:: bash

        salt '*' mount.vfstab
    """
    # NOTE: vfstab is a wrapper for fstab
    return fstab(config)


def rm_fstab(name, device, config="/etc/fstab"):
    """
    .. versionchanged:: 2016.3.2

    Remove the mount point from the fstab

    CLI Example:

    .. code-block:: bash

        salt '*' mount.rm_fstab /mnt/foo /dev/sdg
    """
    modified = False

    if __grains__["kernel"] == "SunOS":
        criteria = _vfstab_entry(name=name, device=device)
    else:
        criteria = _fstab_entry(name=name, device=device)

    lines = []
    try:
        with salt.utils.files.fopen(config, "r") as ifile:
            for line in ifile:
                line = salt.utils.stringutils.to_unicode(line)
                try:
                    if criteria.match(line):
                        modified = True
                    else:
                        lines.append(line)

                except _fstab_entry.ParseError:
                    lines.append(line)
                except _vfstab_entry.ParseError:
                    lines.append(line)

    except OSError as exc:
        msg = "Couldn't read from {0}: {1}"
        raise CommandExecutionError(msg.format(config, exc))

    if modified:
        try:
            with salt.utils.files.fopen(config, "wb") as ofile:
                ofile.writelines(salt.utils.data.encode(lines))
        except OSError as exc:
            msg = "Couldn't write to {0}: {1}"
            raise CommandExecutionError(msg.format(config, exc))

    # Note: not clear why we always return 'True'
    # --just copying previous behavior at this point...
    return True


def rm_vfstab(name, device, config="/etc/vfstab"):
    """
    .. versionadded:: 2016.3.2

    Remove the mount point from the vfstab

    CLI Example:

    .. code-block:: bash

        salt '*' mount.rm_vfstab /mnt/foo /device/c0t0d0p0
    """
    ## NOTE: rm_vfstab is a wrapper for rm_fstab
    return rm_fstab(name, device, config)


def set_fstab(
    name,
    device,
    fstype,
    opts="defaults",
    dump=0,
    pass_num=0,
    config="/etc/fstab",
    test=False,
    match_on="auto",
    not_change=False,
    **kwargs
):
    """
    Verify that this mount is represented in the fstab, change the mount
    to match the data passed, or add the mount if it is not present.

    If the entry is found via `match_on` and `not_change` is True, the
    current line will be preserved.

    CLI Example:

    .. code-block:: bash

        salt '*' mount.set_fstab /mnt/foo /dev/sdz1 ext4
    """

    # Fix the opts type if it is a list
    if isinstance(opts, list):
        opts = ",".join(opts)

    # preserve arguments for updating
    entry_args = {
        "name": name,
        "device": device.replace("\\ ", "\\040"),
        "fstype": fstype,
        "opts": opts.replace("\\ ", "\\040"),
        "dump": dump,
        "pass_num": pass_num,
    }

    lines = []
    ret = None

    # Transform match_on into list--items will be checked later
    if isinstance(match_on, list):
        pass
    elif not isinstance(match_on, str):
        msg = "match_on must be a string or list of strings"
        raise CommandExecutionError(msg)
    elif match_on == "auto":
        # Try to guess right criteria for auto....
        # NOTE: missing some special fstypes here
        specialFSes = frozenset(
            [
                "none",
                "tmpfs",
                "sysfs",
                "proc",
                "fusectl",
                "debugfs",
                "securityfs",
                "devtmpfs",
                "cgroup",
                "nfs",
                "nfs4",
                "glusterfs",
                "btrfs",
            ]
        )

        if fstype in specialFSes:
            match_on = ["name"]
        else:
            match_on = ["device"]
    else:
        match_on = [match_on]

    # generate entry and criteria objects, handle invalid keys in match_on
    entry = _fstab_entry(**entry_args)
    try:
        criteria = entry.pick(match_on)

    except KeyError:
        filterFn = lambda key: key not in _fstab_entry.fstab_keys
        invalid_keys = filter(filterFn, match_on)

        msg = 'Unrecognized keys in match_on: "{}"'.format(invalid_keys)
        raise CommandExecutionError(msg)

    # parse file, use ret to cache status
    if not os.path.isfile(config):
        raise CommandExecutionError('Bad config file "{}"'.format(config))

    try:
        with salt.utils.files.fopen(config, "r") as ifile:
            for line in ifile:
                line = salt.utils.stringutils.to_unicode(line)
                try:
                    if criteria.match(line):
                        # Note: If ret isn't None here,
                        # we've matched multiple lines
                        ret = "present"
                        if entry.match(line) or not_change:
                            lines.append(line)
                        else:
                            ret = "change"
                            lines.append(str(entry))
                    else:
                        lines.append(line)

                except _fstab_entry.ParseError:
                    lines.append(line)

    except OSError as exc:
        msg = "Couldn't read from {0}: {1}"
        raise CommandExecutionError(msg.format(config, exc))

    # add line if not present or changed
    if ret is None:
        lines.append(str(entry))
        ret = "new"

    if ret != "present":  # ret in ['new', 'change']:
        if not salt.utils.args.test_mode(test=test, **kwargs):
            try:
                with salt.utils.files.fopen(config, "wb") as ofile:
                    # The line was changed, commit it!
                    ofile.writelines(salt.utils.data.encode(lines))
            except OSError:
                msg = "File not writable {0}"
                raise CommandExecutionError(msg.format(config))

    return ret


def set_vfstab(
    name,
    device,
    fstype,
    opts="-",
    device_fsck="-",
    pass_fsck="-",
    mount_at_boot="yes",
    config="/etc/vfstab",
    test=False,
    match_on="auto",
    not_change=False,
    **kwargs
):
    """
    .. versionadded:: 2016.3.2

    Verify that this mount is represented in the fstab, change the mount
    to match the data passed, or add the mount if it is not present.

    If the entry is found via `match_on` and `not_change` is True, the
    current line will be preserved.

    CLI Example:

    .. code-block:: bash

        salt '*' mount.set_vfstab /mnt/foo /device/c0t0d0p0 ufs
    """

    # Fix the opts type if it is a list
    if isinstance(opts, list):
        opts = ",".join(opts)

    # Map unknown values for mount_at_boot to no
    if mount_at_boot != "yes":
        mount_at_boot = "no"

    # preserve arguments for updating
    entry_args = {
        "name": name,
        "device": device,
        "fstype": fstype,
        "opts": opts,
        "device_fsck": device_fsck,
        "pass_fsck": pass_fsck,
        "mount_at_boot": mount_at_boot,
    }

    lines = []
    ret = None

    # Transform match_on into list--items will be checked later
    if isinstance(match_on, list):
        pass
    elif not isinstance(match_on, str):
        msg = "match_on must be a string or list of strings"
        raise CommandExecutionError(msg)
    elif match_on == "auto":
        # Try to guess right criteria for auto....
        # NOTE: missing some special fstypes here
        specialFSes = frozenset(
            ["devfs", "proc", "ctfs", "objfs", "sharefs", "fs", "tmpfs"]
        )

        if fstype in specialFSes:
            match_on = ["name"]
        else:
            match_on = ["device"]
    else:
        match_on = [match_on]

    # generate entry and criteria objects, handle invalid keys in match_on
    entry = _vfstab_entry(**entry_args)
    try:
        criteria = entry.pick(match_on)

    except KeyError:
        filterFn = lambda key: key not in _vfstab_entry.vfstab_keys
        invalid_keys = filter(filterFn, match_on)

        msg = 'Unrecognized keys in match_on: "{}"'.format(invalid_keys)
        raise CommandExecutionError(msg)

    # parse file, use ret to cache status
    if not os.path.isfile(config):
        raise CommandExecutionError('Bad config file "{}"'.format(config))

    try:
        with salt.utils.files.fopen(config, "r") as ifile:
            for line in ifile:
                line = salt.utils.stringutils.to_unicode(line)
                try:
                    if criteria.match(line):
                        # Note: If ret isn't None here,
                        # we've matched multiple lines
                        ret = "present"
                        if entry.match(line) or not_change:
                            lines.append(line)
                        else:
                            ret = "change"
                            lines.append(str(entry))
                    else:
                        lines.append(line)

                except _vfstab_entry.ParseError:
                    lines.append(line)

    except OSError as exc:
        msg = "Couldn't read from {0}: {1}"
        raise CommandExecutionError(msg.format(config, exc))

    # add line if not present or changed
    if ret is None:
        lines.append(str(entry))
        ret = "new"

    if ret != "present":  # ret in ['new', 'change']:
        if not salt.utils.args.test_mode(test=test, **kwargs):
            try:
                with salt.utils.files.fopen(config, "wb") as ofile:
                    # The line was changed, commit it!
                    ofile.writelines(salt.utils.data.encode(lines))
            except OSError:
                msg = "File not writable {0}"
                raise CommandExecutionError(msg.format(config))

    return ret


def rm_automaster(name, device, config="/etc/auto_salt"):
    """
    Remove the mount point from the auto_master

    CLI Example:

    .. code-block:: bash

        salt '*' mount.rm_automaster /mnt/foo /dev/sdg
    """
    contents = automaster(config)
    if name not in contents:
        return True
    # The entry is present, get rid of it
    lines = []
    try:
        with salt.utils.files.fopen(config, "r") as ifile:
            for line in ifile:
                line = salt.utils.stringutils.to_unicode(line)
                if line.startswith("#"):
                    # Commented
                    lines.append(line)
                    continue
                if not line.strip():
                    # Blank line
                    lines.append(line)
                    continue
                comps = line.split()
                if len(comps) != 3:
                    # Invalid entry
                    lines.append(line)
                    continue

                comps = line.split()
                prefix = "/.."
                name_chk = comps[0].replace(prefix, "")
                device_fmt = comps[2].split(":")

                if device:
                    if name_chk == name and device_fmt[1] == device:
                        continue
                else:
                    if name_chk == name:
                        continue
                lines.append(line)
    except OSError as exc:
        msg = "Couldn't read from {0}: {1}"
        raise CommandExecutionError(msg.format(config, exc))

    try:
        with salt.utils.files.fopen(config, "wb") as ofile:
            ofile.writelines(salt.utils.data.encode(lines))
    except OSError as exc:
        msg = "Couldn't write to {0}: {1}"
        raise CommandExecutionError(msg.format(config, exc))

    # Update automount
    __salt__["cmd.run"]("automount -cv")
    return True


def set_automaster(
    name,
    device,
    fstype,
    opts="",
    config="/etc/auto_salt",
    test=False,
    not_change=False,
    **kwargs
):
    """
    Verify that this mount is represented in the auto_salt, change the mount
    to match the data passed, or add the mount if it is not present.

    CLI Example:

    .. code-block:: bash

        salt '*' mount.set_automaster /mnt/foo /dev/sdz1 ext4
    """
    # Fix the opts type if it is a list
    if isinstance(opts, list):
        opts = ",".join(opts)
    lines = []
    change = False
    present = False
    automaster_file = "/etc/auto_master"

    if not os.path.isfile(config):
        __salt__["file.touch"](config)
        __salt__["file.append"](automaster_file, "/-\t\t\t{}".format(config))

    name = "/..{}".format(name)
    device_fmt = "{}:{}".format(fstype, device)
    type_opts = "-fstype={},{}".format(fstype, opts)

    if fstype == "smbfs":
        device_fmt = device_fmt.replace(fstype, "")

    try:
        with salt.utils.files.fopen(config, "r") as ifile:
            for line in ifile:
                line = salt.utils.stringutils.to_unicode(line)
                if line.startswith("#"):
                    # Commented
                    lines.append(line)
                    continue
                if not line.strip():
                    # Blank line
                    lines.append(line)
                    continue
                comps = line.split()
                if len(comps) != 3:
                    # Invalid entry
                    lines.append(line)
                    continue
                if comps[0] == name or comps[2] == device_fmt:
                    present = True
                    if not_change:
                        continue
                    # check to see if there are changes
                    # and fix them if there are any
                    if comps[0] != name:
                        change = True
                        comps[0] = name
                    if comps[1] != type_opts:
                        change = True
                        comps[1] = type_opts
                    if comps[2] != device_fmt:
                        change = True
                        comps[2] = device_fmt
                    if change:
                        log.debug(
                            "auto_master entry for mount point %s needs to be updated",
                            name,
                        )
                        newline = "{}\t{}\t{}\n".format(name, type_opts, device_fmt)
                        lines.append(newline)
                else:
                    lines.append(line)
    except OSError as exc:
        msg = "Couldn't read from {0}: {1}"
        raise CommandExecutionError(msg.format(config, exc))

    if change:
        if not salt.utils.args.test_mode(test=test, **kwargs):
            try:
                with salt.utils.files.fopen(config, "wb") as ofile:
                    # The line was changed, commit it!
                    ofile.writelines(salt.utils.data.encode(lines))
            except OSError:
                msg = "File not writable {0}"
                raise CommandExecutionError(msg.format(config))

        return "change"

    if not change:
        if present:
            # The right entry is already here
            return "present"
        else:
            if not salt.utils.args.test_mode(test=test, **kwargs):
                # The entry is new, add it to the end of the fstab
                newline = "{}\t{}\t{}\n".format(name, type_opts, device_fmt)
                lines.append(newline)
                try:
                    with salt.utils.files.fopen(config, "wb") as ofile:
                        # The line was changed, commit it!
                        ofile.writelines(salt.utils.data.encode(lines))
                except OSError:
                    raise CommandExecutionError("File not writable {}".format(config))
    return "new"


def automaster(config="/etc/auto_salt"):
    """
    List the contents of the auto master

    CLI Example:

    .. code-block:: bash

        salt '*' mount.automaster
    """
    ret = {}
    if not os.path.isfile(config):
        return ret
    with salt.utils.files.fopen(config) as ifile:
        for line in ifile:
            line = salt.utils.stringutils.to_unicode(line)
            if line.startswith("#"):
                # Commented
                continue
            if not line.strip():
                # Blank line
                continue
            comps = line.split()
            if len(comps) != 3:
                # Invalid entry
                continue

            prefix = "/.."
            name = comps[0].replace(prefix, "")
            device_fmt = comps[2].split(":")
            opts = comps[1].split(",")

            ret[name] = {"device": device_fmt[1], "fstype": opts[0], "opts": opts[1:]}
    return ret


def mount(
    name, device=False, mkmnt=False, fstype="", opts="defaults", user=None, util="mount"
):
    """
    Mount a device

    CLI Example:

    .. code-block:: bash

        salt '*' mount.mount /mnt/foo /dev/sdz1 True
    """
    if util != "mount":
        # This functionality used to live in img.mount_image
        if util == "guestfs":
            return __salt__["guestfs.mount"](name, root=device)
        elif util == "qemu_nbd":
            mnt = __salt__["qemu_nbd.init"](name, device)
            if not mnt:
                return False
            first = next(iter(mnt.keys()))
            __context__["img.mnt_{}".format(first)] = mnt
            return first
        return False

    # Darwin doesn't expect defaults when mounting without other options
    if "defaults" in opts and __grains__["os"] in ["MacOS", "Darwin", "AIX"]:
        opts = None

    if isinstance(opts, str):
        opts = opts.split(",")

    if not os.path.exists(name) and mkmnt:
        __salt__["file.mkdir"](name, user=user)

    args = ""
    if opts is not None:
        lopts = ",".join(opts)
        args = "-o {}".format(lopts)

    if fstype:
        # use of fstype on AIX differs from typical Linux use of -t
        # functionality AIX uses -v vfsname, -t fstype mounts all with
        # fstype in /etc/filesystems
        if "AIX" in __grains__["os"]:
            args += " -v {}".format(fstype)
        elif "solaris" in __grains__["os"].lower():
            args += " -F {}".format(fstype)
        else:
            args += " -t {}".format(fstype)

    cmd = "mount "
    if device:
        cmd += "{} {} {} ".format(args, device, name)
    else:
        cmd += "{} ".format(name)
    out = __salt__["cmd.run_all"](cmd, runas=user, python_shell=False)
    if out["retcode"]:
        return out["stderr"]
    return True


def remount(name, device, mkmnt=False, fstype="", opts="defaults", user=None):
    """
    Attempt to remount a device, if the device is not already mounted, mount
    is called

    CLI Example:

    .. code-block:: bash

        salt '*' mount.remount /mnt/foo /dev/sdz1 True
    """
    force_mount = False
    if __grains__["os"] in ["MacOS", "Darwin"]:
        if opts == "defaults":
            opts = "noowners"
        if fstype == "smbfs":
            force_mount = True

    if "AIX" in __grains__["os"]:
        if opts == "defaults":
            opts = []

    if isinstance(opts, str):
        opts = opts.split(",")
    mnts = active()
    if name in mnts:
        # The mount point is mounted, attempt to remount it with the given data
        if "remount" not in opts and __grains__["os"] not in [
            "OpenBSD",
            "MacOS",
            "Darwin",
        ]:
            opts.append("remount")
        if force_mount:
            # We need to force the mount but first we should unmount
            umount(name, device, user=user)
        lopts = ",".join(opts)
        args = "-o {}".format(lopts)

        if fstype:
            # use of fstype on AIX differs from typical Linux use of
            # -t functionality AIX uses -v vfsname, -t fstype mounts
            # all with fstype in /etc/filesystems
            if "AIX" in __grains__["os"]:
                args += " -v {}".format(fstype)
            elif "solaris" in __grains__["os"].lower():
                args += " -F {}".format(fstype)
            else:
                args += " -t {}".format(fstype)

        if __grains__["os"] not in ["OpenBSD", "MacOS", "Darwin"] or force_mount:
            cmd = "mount {} {} {} ".format(args, device, name)
        else:
            cmd = "mount -u {} {} {} ".format(args, device, name)
        out = __salt__["cmd.run_all"](cmd, runas=user, python_shell=False)
        if out["retcode"]:
            return out["stderr"]
        return True
    # Mount a filesystem that isn't already
    return mount(name, device, mkmnt, fstype, opts, user=user)


def umount(name, device=None, user=None, util="mount"):
    """
    Attempt to unmount a device by specifying the directory it is mounted on

    CLI Example:

    .. code-block:: bash

        salt '*' mount.umount /mnt/foo

    .. versionadded:: 2015.5.0

    .. code-block:: bash

        salt '*' mount.umount /mnt/foo /dev/xvdc1
    """
    if util == "guestfs":
        __salt__["guestfs.umount"](name, disk=device)
        return
    elif util == "qemu_nbd":
        # This functionality used to live in img.umount_image
        if "qemu_nbd.clear" in __salt__:
            if "img.mnt_{}".format(name) in __context__:
                __salt__["qemu_nbd.clear"](__context__["img.mnt_{}".format(name)])
                return

    mnts = active()
    if name not in mnts:
        return "{} does not have anything mounted".format(name)

    if not device:
        cmd = "umount {}".format(name)
    else:
        cmd = "umount {}".format(device)
    out = __salt__["cmd.run_all"](cmd, runas=user, python_shell=False)
    if out["retcode"]:
        return out["stderr"]
    return True


def is_fuse_exec(cmd):
    """
    Returns true if the command passed is a fuse mountable application.

    CLI Example:

    .. code-block:: bash

        salt '*' mount.is_fuse_exec sshfs
    """
    cmd_path = salt.utils.path.which(cmd)

    # No point in running ldd on a command that doesn't exist
    if not cmd_path:
        return False
    elif not salt.utils.path.which("ldd"):
        raise CommandNotFoundError("ldd")

    out = __salt__["cmd.run"]("ldd {}".format(cmd_path), python_shell=False)
    return "libfuse" in out


def swaps():
    """
    Return a dict containing information on active swap

    .. versionchanged:: 2016.3.2

    CLI Example:

    .. code-block:: bash

        salt '*' mount.swaps
    """
    ret = {}
    if __grains__["kernel"] == "SunOS":
        for line in __salt__["cmd.run_stdout"]("swap -l").splitlines():
            if line.startswith("swapfile"):
                continue
            comps = line.split()
            ret[comps[0]] = {
                "type": "device" if comps[0].startswith(("/dev", "swap")) else "file",
                "size": int(comps[3]),
                "used": (int(comps[3]) - int(comps[4])),
                "priority": "-",
            }
    elif "AIX" in __grains__["kernel"]:
        for line in __salt__["cmd.run_stdout"]("swap -l").splitlines():
            if line.startswith("device"):
                continue
            comps = line.split()

            # AIX uses MB for units
            ret[comps[0]] = {
                "type": "device",
                "size": int(comps[3][:-2]) * 1024,
                "used": (int(comps[3][:-2]) - int(comps[4][:-2])) * 1024,
                "priority": "-",
            }
    elif __grains__["os"] != "OpenBSD":
        with salt.utils.files.fopen("/proc/swaps") as fp_:
            for line in fp_:
                line = salt.utils.stringutils.to_unicode(line)
                if line.startswith("Filename"):
                    continue
                comps = line.split()
                ret[comps[0]] = {
                    "type": comps[1],
                    "size": comps[2],
                    "used": comps[3],
                    "priority": comps[4],
                }
    else:
        for line in __salt__["cmd.run_stdout"]("swapctl -kl").splitlines():
            if line.startswith(("Device", "Total")):
                continue
            swap_type = "file"
            comps = line.split()
            if comps[0].startswith("/dev/"):
                swap_type = "partition"
            ret[comps[0]] = {
                "type": swap_type,
                "size": comps[1],
                "used": comps[2],
                "priority": comps[5],
            }
    return ret


def swapon(name, priority=None):
    """
    Activate a swap disk

    .. versionchanged:: 2016.3.2

    CLI Example:

    .. code-block:: bash

        salt '*' mount.swapon /root/swapfile
    """
    ret = {}
    on_ = swaps()
    if name in on_:
        ret["stats"] = on_[name]
        ret["new"] = False
        return ret

    if __grains__["kernel"] == "SunOS":
        if __grains__["virtual"] != "zone":
            __salt__["cmd.run"]("swap -a {}".format(name), python_shell=False)
        else:
            return False
    else:
        cmd = "swapon {}".format(name)
        if priority and "AIX" not in __grains__["kernel"]:
            cmd += " -p {}".format(priority)
        __salt__["cmd.run"](cmd, python_shell=False)

    on_ = swaps()
    if name in on_:
        ret["stats"] = on_[name]
        ret["new"] = True
        return ret

    return ret


def swapoff(name):
    """
    Deactivate a named swap mount

    .. versionchanged:: 2016.3.2

    CLI Example:

    .. code-block:: bash

        salt '*' mount.swapoff /root/swapfile
    """
    on_ = swaps()
    if name in on_:
        if __grains__["kernel"] == "SunOS":
            if __grains__["virtual"] != "zone":
                __salt__["cmd.run"]("swap -a {}".format(name), python_shell=False)
            else:
                return False
        elif __grains__["os"] != "OpenBSD":
            __salt__["cmd.run"]("swapoff {}".format(name), python_shell=False)
        else:
            __salt__["cmd.run"]("swapctl -d {}".format(name), python_shell=False)
        on_ = swaps()
        if name in on_:
            return False
        return True
    return None


def is_mounted(name):
    """
    .. versionadded:: 2014.7.0

    Provide information if the path is mounted

    CLI Example:

    .. code-block:: bash

        salt '*' mount.is_mounted /mnt/share
    """
    active_ = active()
    if name in active_:
        return True
    else:
        return False


def read_mount_cache(name):
    """
    .. versionadded:: 2018.3.0

    Provide information if the path is mounted

    CLI Example:

    .. code-block:: bash

        salt '*' mount.read_mount_cache /mnt/share
    """
    cache = salt.utils.mount.read_cache(__opts__)
    if cache:
        if "mounts" in cache and cache["mounts"]:
            if name in cache["mounts"]:
                return cache["mounts"][name]
    return {}


def write_mount_cache(real_name, device, mkmnt, fstype, mount_opts):
    """
    .. versionadded:: 2018.3.0

    Provide information if the path is mounted

    :param real_name:     The real name of the mount point where the device is mounted.
    :param device:        The device that is being mounted.
    :param mkmnt:         Whether or not the mount point should be created.
    :param fstype:        The file system that is used.
    :param mount_opts:    Additional options used when mounting the device.
    :return:              Boolean if message was sent successfully.

    CLI Example:

    .. code-block:: bash

        salt '*' mount.write_mount_cache /mnt/share /dev/sda1 False ext4 defaults,nosuid
    """
    cache = salt.utils.mount.read_cache(__opts__)

    if not cache:
        cache = {}
        cache["mounts"] = {}
    else:
        if "mounts" not in cache:
            cache["mounts"] = {}

    cache["mounts"][real_name] = {
        "device": device,
        "fstype": fstype,
        "mkmnt": mkmnt,
        "opts": mount_opts,
    }

    cache_write = salt.utils.mount.write_cache(cache, __opts__)
    if cache_write:
        return True
    else:
        raise CommandExecutionError("Unable to write mount cache.")


def delete_mount_cache(real_name):
    """
    .. versionadded:: 2018.3.0

    Provide information if the path is mounted

    CLI Example:

    .. code-block:: bash

        salt '*' mount.delete_mount_cache /mnt/share
    """
    cache = salt.utils.mount.read_cache(__opts__)

    if cache:
        if "mounts" in cache:
            if real_name in cache["mounts"]:
                del cache["mounts"][real_name]
                cache_write = salt.utils.mount.write_cache(cache, __opts__)
                if not cache_write:
                    raise CommandExecutionError("Unable to write mount cache.")
    return True


def _filesystems(config="/etc/filesystems", leading_key=True):
    """
    Return the contents of the filesystems in an OrderedDict

    config
        File containing filesystem infomation

    leading_key
        True    return dictionary keyed by 'name' and value as dictionary with other keys, values (name excluded)
                OrderedDict({ '/dir' : OrderedDict({'dev': '/dev/hd8', .... }}))

        False   return dictionary  keyed by 'name' and value as dictionary with all keys, values (name included)
                OrderedDict({ '/dir' : OrderedDict({'name': '/dir', 'dev': '/dev/hd8', ... })})
    """
    ret = OrderedDict()
    lines = []
    parsing_block = False
    if not os.path.isfile(config) or "AIX" not in __grains__["kernel"]:
        return ret

    # read in block of filesystems, block starts with '/' till empty line
    with salt.utils.files.fopen(config) as ifile:
        for line in ifile:
            line = salt.utils.stringutils.to_unicode(line)

            # skip till first entry
            if not line.startswith("/") and not parsing_block:
                continue

            if line.startswith("/"):
                parsing_block = True
                lines.append(line)
            elif not line.split():
                parsing_block = False
                try:
                    entry = _FileSystemsEntry.dict_from_lines(
                        lines, _FileSystemsEntry.compatibility_keys
                    )
                    lines = []
                    if "opts" in entry:
                        entry["opts"] = entry["opts"].split(",")
                    while entry["name"] in ret:
                        entry["name"] += "_"

                    if leading_key:
                        ret[entry.pop("name")] = entry
                    else:
                        ret[entry["name"]] = entry

                except _FileSystemsEntry.ParseError:
                    pass
            else:
                lines.append(line)

    return ret


def filesystems(config="/etc/filesystems"):
    """
    .. versionadded:: 2018.3.3

    List the contents of the filesystems

    CLI Example:

    .. code-block:: bash

        salt '*' mount.filesystems
    """
    ret = {}
    if "AIX" not in __grains__["kernel"]:
        return ret

    ret_dict = _filesystems(config)
    if ret_dict:
        ret_key = next(iter(ret_dict.keys()))
        ret = {ret_key: dict(ret_dict[ret_key])}

    return ret


def set_filesystems(
    name,
    device,
    vfstype,
    opts="-",
    mount="true",
    config="/etc/filesystems",
    test=False,
    match_on="auto",
    not_change=False,
    **kwargs
):
    """
    .. versionadded:: 2018.3.3

    Verify that this mount is represented in the filesystems, change the mount
    to match the data passed, or add the mount if it is not present on AIX

    If the entry is found via `match_on` and `not_change` is True, the
    current line will be preserved.

        Provide information if the path is mounted

    :param name:          The name of the mount point where the device is mounted.
    :param device:        The device that is being mounted.
    :param vfstype:       The file system that is used (AIX has two fstypes, fstype and vfstype - similar to Linux fstype)
    :param opts:          Additional options used when mounting the device.
    :param mount:         Mount if not mounted, default True.
    :param config:        Configuration file, default /etc/filesystems.
    :param match:         File systems type to match on, default auto

    CLI Example:

    .. code-block:: bash

        salt '*' mount.set_filesystems /mnt/foo /dev/sdz1 jfs2
    """
    # Fix the opts type if it is a list
    if isinstance(opts, list):
        opts = ",".join(opts)

    # preserve arguments for updating
    entry_args = {
        "name": name,
        "dev": device.replace("\\ ", "\\040"),
        "vfstype": vfstype,
        "opts": opts,
        "mount": mount,
    }

    view_lines = []
    ret = None

    if "AIX" not in __grains__["kernel"]:
        return ret

    # Transform match_on into list--items will be checked later
    if isinstance(match_on, list):
        pass
    elif not isinstance(match_on, str):
        raise CommandExecutionError("match_on must be a string or list of strings")
    elif match_on == "auto":
        # Try to guess right criteria for auto....
        # added IBM types from sys/vmount.h after btrfs
        # NOTE: missing some special fstypes here
        specialFSes = frozenset(
            [
                "none",
                "tmpfs",
                "sysfs",
                "proc",
                "fusectl",
                "debugfs",
                "securityfs",
                "devtmpfs",
                "cgroup",
                "btrfs",
                "cdrfs",
                "procfs",
                "jfs",
                "jfs2",
                "nfs",
                "sfs",
                "nfs3",
                "cachefs",
                "udfs",
                "cifs",
                "namefs",
                "pmemfs",
                "ahafs",
                "nfs4",
                "autofs",
                "stnfs",
            ]
        )

        if vfstype in specialFSes:
            match_on = ["name"]
        else:
            match_on = ["dev"]
    else:
        match_on = [match_on]

    # generate entry and criteria objects, handle invalid keys in match_on
    entry_ip = _FileSystemsEntry.from_line(entry_args, kwargs)
    try:
        criteria = entry_ip.pick(match_on)

    except KeyError:
        filterFn = lambda key: key not in _FileSystemsEntry.compatibility_keys
        invalid_keys = filter(filterFn, match_on)
        raise CommandExecutionError(
            'Unrecognized keys in match_on: "{}"'.format(invalid_keys)
        )

    # parse file, use ret to cache status
    if not os.path.isfile(config):
        raise CommandExecutionError('Bad config file "{}"'.format(config))

    # read in block of filesystem, block starts with '/' till empty line
    try:
        fsys_filedict = _filesystems(config, False)
        for fsys_view in fsys_filedict.items():
            if criteria.match(fsys_view):
                ret = "present"
                if entry_ip.match(fsys_view) or not_change:
                    view_lines.append(fsys_view)
                else:
                    ret = "change"
                    kv = entry_ip["name"]
                    view_lines.append((kv, entry_ip))
            else:
                view_lines.append(fsys_view)

    except OSError as exc:
        raise CommandExecutionError("Couldn't read from {}: {}".format(config, exc))

    # add line if not present or changed
    if ret is None:
        for dict_view in entry_ip.dict_from_entry().items():
            view_lines.append(dict_view)
        ret = "new"

    if ret != "present":  # ret in ['new', 'change']:
        try:
            with salt.utils.files.fopen(config, "wb") as ofile:
                # The line was changed, commit it!
                for fsys_view in view_lines:
                    entry = fsys_view[1]
                    mystrg = _FileSystemsEntry.dict_to_lines(entry)
                    ofile.writelines(salt.utils.data.encode(mystrg))
        except OSError:
            raise CommandExecutionError("File not writable {}".format(config))

    return ret


def rm_filesystems(name, device, config="/etc/filesystems"):
    """
    .. versionadded:: 2018.3.3

    Remove the mount point from the filesystems

    CLI Example:

    .. code-block:: bash

        salt '*' mount.rm_filesystems /mnt/foo /dev/sdg
    """
    modified = False
    view_lines = []

    if "AIX" not in __grains__["kernel"]:
        return modified

    criteria = _FileSystemsEntry(name=name, dev=device)
    try:
        fsys_filedict = _filesystems(config, False)
        for fsys_view in fsys_filedict.items():
            try:
                if criteria.match(fsys_view):
                    modified = True
                else:
                    view_lines.append(fsys_view)

            except _FileSystemsEntry.ParseError:
                view_lines.append(fsys_view)

    except OSError as exc:
        raise CommandExecutionError("Couldn't read from {}: {}".format(config, exc))

    if modified:
        try:
            with salt.utils.files.fopen(config, "wb") as ofile:
                for fsys_view in view_lines:
                    entry = fsys_view[1]
                    mystrg = _FileSystemsEntry.dict_to_lines(entry)
                    ofile.writelines(salt.utils.data.encode(mystrg))
        except OSError as exc:
            raise CommandExecutionError("Couldn't write to {}: {}".format(config, exc))

    return modified



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import salt.modules.riak as riak
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import patch
from tests.support.unit import TestCase


class RiakTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.modules.riak
    """

    def setup_loader_modules(self):
        return {riak: {}}

    def test_start(self):
        """
        Test for start Riak
        """
        with patch.object(
            riak, "__execute_cmd", return_value={"retcode": 0, "stdout": "success"}
        ):
            self.assertEqual(riak.start(), {"success": True, "comment": "success"})

    def test_stop(self):
        """
        Test for stop Riak
        """
        with patch.object(
            riak, "__execute_cmd", return_value={"retcode": 0, "stdout": "success"}
        ):
            self.assertEqual(riak.stop(), {"success": True, "comment": "success"})

    def test_cluster_join(self):
        """
        Test for Join a Riak cluster
        """
        with patch.object(
            riak, "__execute_cmd", return_value={"retcode": 0, "stdout": "success"}
        ):
            self.assertEqual(
                riak.cluster_join("A", "B"), {"success": True, "comment": "success"}
            )

    def test_cluster_leave(self):
        """
        Test for leaving a Riak cluster
        """
        with patch.object(
            riak, "__execute_cmd", return_value={"retcode": 0, "stdout": "success"}
        ):
            self.assertEqual(
                riak.cluster_leave("A", "B"), {"success": True, "comment": "success"}
            )

    def test_cluster_plan(self):
        """
        Test for Review Cluster Plan
        """
        with patch.object(
            riak, "__execute_cmd", return_value={"retcode": 0, "stdout": "success"}
        ):
            self.assertTrue(riak.cluster_plan())

    def test_cluster_commit(self):
        """
        Test for Commit Cluster Changes
        """
        with patch.object(
            riak, "__execute_cmd", return_value={"retcode": 0, "stdout": "success"}
        ):
            self.assertEqual(
                riak.cluster_commit(), {"success": True, "comment": "success"}
            )

    def test_member_status(self):
        """
        Test for Get cluster member status
        """
        with patch.object(
            riak, "__execute_cmd", return_value={"stdout": "A:a/B:b\nC:c/D:d"}
        ):
            self.assertDictEqual(
                riak.member_status(),
                {
                    "membership": {},
                    "summary": {
                        "A": "a",
                        "C": "c",
                        "B": "b",
                        "D": "d",
                        "Exiting": 0,
                        "Down": 0,
                        "Valid": 0,
                        "Leaving": 0,
                        "Joining": 0,
                    },
                },
            )

    def test_status(self):
        """
        Test status information
        """
        ret = {"stdout": "vnode_map_update_time_95 : 0\nvnode_map_update_time_99 : 0"}

        with patch.object(riak, "__execute_cmd", return_value=ret):
            self.assertEqual(
                riak.status(),
                {"vnode_map_update_time_95": "0", "vnode_map_update_time_99": "0"},
            )

    def test_test(self):
        """
        Test the Riak test
        """
        with patch.object(
            riak, "__execute_cmd", return_value={"retcode": 0, "stdout": "success"}
        ):
            self.assertEqual(riak.test(), {"success": True, "comment": "success"})

    def test_services(self):
        """
        Test Riak Service List
        """
        with patch.object(riak, "__execute_cmd", return_value={"stdout": "[a,b,c]"}):
            self.assertEqual(riak.services(), ["a", "b", "c"])



"""
Modules used to control the master itself
"""

from collections.abc import Mapping

import salt.client.mixins
import salt.config
import salt.loader
import salt.transport.client
import salt.utils.error
import salt.utils.zeromq


class WheelClient(
    salt.client.mixins.SyncClientMixin, salt.client.mixins.AsyncClientMixin
):
    """
    An interface to Salt's wheel modules

    :ref:`Wheel modules <all-salt.wheel>` interact with various parts of the
    Salt Master.

    Importing and using ``WheelClient`` must be done on the same machine as the
    Salt Master and it must be done using the same user that the Salt Master is
    running as. Unless :conf_master:`external_auth` is configured and the user
    is authorized to execute wheel functions: (``@wheel``).

    Usage:

    .. code-block:: python

        import salt.config
        import salt.wheel
        opts = salt.config.master_config('/etc/salt/master')
        wheel = salt.wheel.WheelClient(opts)
    """

    client = "wheel"
    tag_prefix = "wheel"

    def __init__(self, opts=None):
        self.opts = opts
        self.context = {}
        self.functions = salt.loader.wheels(opts, context=self.context)

    # TODO: remove/deprecate
    def call_func(self, fun, **kwargs):
        """
        Backwards compatibility
        """
        return self.low(
            fun,
            kwargs,
            print_event=kwargs.get("print_event", True),
            full_return=kwargs.get("full_return", False),
        )

    # TODO: Inconsistent with runner client-- the runner client's master_call gives
    # an asynchronous return, unlike this
    def master_call(self, **kwargs):
        """
        Execute a wheel function through the master network interface (eauth).
        """
        load = kwargs
        load["cmd"] = "wheel"
        interface = self.opts["interface"]
        if interface == "0.0.0.0":
            interface = "127.0.0.1"
        master_uri = "tcp://{}:{}".format(
            salt.utils.zeromq.ip_bracket(interface),
            str(self.opts["ret_port"]),
        )
        with salt.transport.client.ReqChannel.factory(
            self.opts, crypt="clear", master_uri=master_uri, usage="master_call"
        ) as channel:
            ret = channel.send(load)

        if isinstance(ret, Mapping):
            if "error" in ret:
                salt.utils.error.raise_error(**ret["error"])

        return ret

    def cmd_sync(self, low, timeout=None, full_return=False):
        """
        Execute a wheel function synchronously; eauth is respected

        This function requires that :conf_master:`external_auth` is configured
        and the user is authorized to execute runner functions: (``@wheel``).

        .. code-block:: python

            >>> wheel.cmd_sync({
            'fun': 'key.finger',
            'match': 'jerry',
            'eauth': 'auto',
            'username': 'saltdev',
            'password': 'saltdev',
            })
            {'minions': {'jerry': '5d:f6:79:43:5e:d4:42:3f:57:b8:45:a8:7e:a4:6e:ca'}}
        """
        return self.master_call(**low)

    # TODO: Inconsistent with runner client-- that one uses the master_call function
    # and runs within the master daemon. Need to pick one...
    def cmd_async(self, low):
        """
        Execute a function asynchronously; eauth is respected

        This function requires that :conf_master:`external_auth` is configured
        and the user is authorized

        .. code-block:: python

            >>> wheel.cmd_async({
                'fun': 'key.finger',
                'match': 'jerry',
                'eauth': 'auto',
                'username': 'saltdev',
                'password': 'saltdev',
            })
            {'jid': '20131219224744416681', 'tag': 'salt/wheel/20131219224744416681'}
        """
        fun = low.get("fun")
        return self.asynchronous(fun, low, local=False)

    def cmd(
        self,
        fun,
        arg=None,
        pub_data=None,
        kwarg=None,
        print_event=True,
        full_return=False,
    ):  # pylint: disable=useless-super-delegation
        """
        Execute a function

        .. code-block:: python

            >>> wheel.cmd('key.finger', ['jerry'])
            {'minions': {'jerry': '5d:f6:79:43:5e:d4:42:3f:57:b8:45:a8:7e:a4:6e:ca'}}
        """
        return super().cmd(fun, arg, pub_data, kwarg, print_event, full_return)


Wheel = WheelClient  # for backward-compat



import salt.modules.pillar as pillarmod
from salt.utils.odict import OrderedDict
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase

pillar_value_1 = dict(a=1, b="very secret")


class PillarModuleTestCase(TestCase, LoaderModuleMockMixin):
    def setup_loader_modules(self):
        return {pillarmod: {}}

    def test_obfuscate_inner_recursion(self):
        self.assertEqual(
            pillarmod._obfuscate_inner(
                dict(a=[1, 2], b=dict(pwd="secret", deeper=("a", 1)))
            ),
            dict(a=["<int>", "<int>"], b=dict(pwd="<str>", deeper=("<str>", "<int>"))),
        )

    def test_obfuscate_inner_more_types(self):
        self.assertEqual(
            pillarmod._obfuscate_inner(OrderedDict([("key", "value")])),
            OrderedDict([("key", "<str>")]),
        )

        self.assertEqual(pillarmod._obfuscate_inner({1, 2}), {"<int>"})

        self.assertEqual(pillarmod._obfuscate_inner((1, 2)), ("<int>", "<int>"))

    def test_obfuscate(self):
        with patch("salt.modules.pillar.items", MagicMock(return_value=pillar_value_1)):
            self.assertEqual(pillarmod.obfuscate(), dict(a="<int>", b="<str>"))

    def test_ls(self):
        with patch("salt.modules.pillar.items", MagicMock(return_value=pillar_value_1)):
            ls = sorted(pillarmod.ls())
            self.assertCountEqual(ls, ["a", "b"])

    def test_pillar_get_default_merge(self):
        defaults = {
            "int": 1,
            "string": "foo",
            "list": ["foo"],
            "dict": {"foo": "bar", "subkey": {"foo": "bar"}},
        }

        pillar_mock = {
            "int": 2,
            "string": "bar",
            "list": ["bar", "baz"],
            "dict": {"baz": "qux", "subkey": {"baz": "qux"}},
        }

        # Test that we raise a KeyError when pillar_raise_on_missing is True
        with patch.dict(pillarmod.__opts__, {"pillar_raise_on_missing": True}):
            self.assertRaises(KeyError, pillarmod.get, "missing")
        # Test that we return an empty string when it is not
        with patch.dict(pillarmod.__opts__, {}):
            self.assertEqual(pillarmod.get("missing"), "")

        with patch.dict(pillarmod.__pillar__, pillar_mock):
            # Test with no default passed (it should be KeyError) and merge=True.
            # The merge should be skipped and the value returned from __pillar__
            # should be returned.
            for item in pillarmod.__pillar__:
                self.assertEqual(
                    pillarmod.get(item, merge=True), pillarmod.__pillar__[item]
                )

            # Test merging when the type of the default value is not the same as
            # what was returned. Merging should be skipped and the value returned
            # from __pillar__ should be returned.
            for default_type in defaults:
                for data_type in ("dict", "list"):
                    if default_type == data_type:
                        continue
                    self.assertEqual(
                        pillarmod.get(
                            data_type, default=defaults[default_type], merge=True
                        ),
                        pillarmod.__pillar__[data_type],
                    )

            # Test recursive dict merging
            self.assertEqual(
                pillarmod.get("dict", default=defaults["dict"], merge=True),
                {"foo": "bar", "baz": "qux", "subkey": {"foo": "bar", "baz": "qux"}},
            )

            # Test list merging
            self.assertEqual(
                pillarmod.get("list", default=defaults["list"], merge=True),
                ["foo", "bar", "baz"],
            )

    def test_pillar_get_default_merge_regression_38558(self):
        """Test for pillar.get(key=..., default=..., merge=True)

        Do not update the ``default`` value when using ``merge=True``.

        See: https://github.com/saltstack/salt/issues/38558
        """
        with patch.dict(pillarmod.__pillar__, {"l1": {"l2": {"l3": 42}}}):

            res = pillarmod.get(key="l1")
            self.assertEqual({"l2": {"l3": 42}}, res)

            default = {"l2": {"l3": 43}}

            res = pillarmod.get(key="l1", default=default)
            self.assertEqual({"l2": {"l3": 42}}, res)
            self.assertEqual({"l2": {"l3": 43}}, default)

            res = pillarmod.get(key="l1", default=default, merge=True)
            self.assertEqual({"l2": {"l3": 42}}, res)
            self.assertEqual({"l2": {"l3": 43}}, default)

    def test_pillar_get_default_merge_regression_39062(self):
        """
        Confirm that we do not raise an exception if default is None and
        merge=True.

        See https://github.com/saltstack/salt/issues/39062 for more info.
        """
        with patch.dict(pillarmod.__pillar__, {"foo": "bar"}):

            self.assertEqual(
                pillarmod.get(key="foo", default=None, merge=True),
                "bar",
            )

    def test_pillar_get_int_key(self):
        """
        Confirm that we can access pillar keys that are integers
        """
        with patch.dict(pillarmod.__pillar__, {12345: "luggage_code"}):

            self.assertEqual(
                pillarmod.get(key=12345, default=None, merge=True),
                "luggage_code",
            )

        with patch.dict(
            pillarmod.__pillar__, {12345: {"l2": {"l3": "my_luggage_code"}}}
        ):

            res = pillarmod.get(key=12345)
            self.assertEqual({"l2": {"l3": "my_luggage_code"}}, res)

            default = {"l2": {"l3": "your_luggage_code"}}

            res = pillarmod.get(key=12345, default=default)
            self.assertEqual({"l2": {"l3": "my_luggage_code"}}, res)
            self.assertEqual({"l2": {"l3": "your_luggage_code"}}, default)

            res = pillarmod.get(key=12345, default=default, merge=True)
            self.assertEqual({"l2": {"l3": "my_luggage_code"}}, res)
            self.assertEqual({"l2": {"l3": "your_luggage_code"}}, default)



import logging
import os.path
from collections import namedtuple

import pkg_resources  # pylint: disable=3rd-party-module-not-gated

import salt.config
import salt.loader
import salt.utils.versions
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.runtests import RUNTIME_VARS
from tests.support.unit import TestCase, skipIf

# pylint: disable=import-error
try:
    import salt.modules.boto_route53 as boto_route53
    from boto.route53.exception import DNSServerError
    import boto

    boto.ENDPOINTS_PATH = os.path.join(
        RUNTIME_VARS.TESTS_DIR, "unit/files/endpoints.json"
    )
    from moto import mock_route53_deprecated

    HAS_MOTO = True
except ImportError:
    HAS_MOTO = False

    def mock_route53_deprecated(self):
        """
        if the mock_route53_deprecated function is not available due to import failure
        this replaces the decorated function with stub_function.
        Allows boto_route53 unit tests to use the @mock_route53_deprecated decorator
        without a "NameError: name 'mock_route53_deprecated' is not defined" error.
        """

        def stub_function(self):
            pass

        return stub_function


# pylint: enable=import-error

log = logging.getLogger(__name__)

required_moto = "1.0.1"


def _has_required_moto():
    """
    Returns True or False depending on if ``moto`` is installed and at the correct version,
    depending on what version of Python is running these tests.
    """
    if not HAS_MOTO:
        return False
    else:
        moto_version = salt.utils.versions.LooseVersion(
            pkg_resources.get_distribution("moto").version
        )
        if moto_version < salt.utils.versions.LooseVersion(required_moto):
            return False
    return True


@skipIf(HAS_MOTO is False, "The moto module must be installed.")
@skipIf(
    _has_required_moto() is False,
    "The moto module must be >= to {}".format(required_moto),
)
class BotoRoute53TestCase(TestCase, LoaderModuleMockMixin):
    """
    TestCase for salt.modules.boto_route53 module
    """

    def setup_loader_modules(self):
        self.opts = salt.config.DEFAULT_MINION_OPTS.copy()
        self.opts["route53.keyid"] = "GKTADJGHEIQSXMKKRBJ08H"
        self.opts["route53.key"] = "askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs"
        utils = salt.loader.utils(self.opts)
        funcs = salt.loader.minion_mods(
            self.opts, utils=utils, whitelist=["boto_route53", "config"]
        )
        return {
            boto_route53: {
                "__opts__": self.opts,
                "__utils__": utils,
                "__salt__": funcs,
            },
        }

    def setUp(self):
        TestCase.setUp(self)
        # __virtual__ must be caller in order for _get_conn to be injected
        boto_route53.__virtual__()
        boto_route53.__init__(self.opts)

    def tearDown(self):
        del self.opts

    @mock_route53_deprecated
    def test_create_healthcheck(self):
        """
        tests that given a valid instance id and valid ELB that
        register_instances returns True.
        """
        expected = {
            "result": {
                "CreateHealthCheckResponse": {
                    "HealthCheck": {
                        "HealthCheckConfig": {
                            "FailureThreshold": "3",
                            "IPAddress": "10.0.0.1",
                            "ResourcePath": "/",
                            "RequestInterval": "30",
                            "Type": "HTTPS",
                            "Port": "443",
                            "FullyQualifiedDomainName": "blog.saltstack.furniture",
                        },
                        "HealthCheckVersion": "1",
                    },
                },
            },
        }
        healthcheck = boto_route53.create_healthcheck(
            "10.0.0.1",
            fqdn="blog.saltstack.furniture",
            hc_type="HTTPS",
            port=443,
            resource_path="/",
        )
        del healthcheck["result"]["CreateHealthCheckResponse"]["HealthCheck"][
            "CallerReference"
        ]
        del healthcheck["result"]["CreateHealthCheckResponse"]["HealthCheck"]["Id"]
        self.assertEqual(healthcheck, expected)


class DummyConn:
    """
    Simple object housing a mock to simulate Error conditions. Each keyword
    argument passed into this will be set as MagicMock with the keyword value
    being set as the side_effect for that function.
    """

    def __init__(self, **kwargs):
        for key, val in kwargs.items():
            setattr(self, key, MagicMock(side_effect=val))


class BotoRoute53RetryTestCase(TestCase, LoaderModuleMockMixin):
    """
    TestCase for salt.modules.boto_route53 module
    """

    _retryable_error = DNSServerError(
        555, "Rejected", body={"Error": {"Code": "Foo", "Message": "Bar"}}
    )
    _fatal_error = DNSServerError(
        666,
        "Flagrant System Error",
        body={
            "Error": {
                "Code": "SignatureDoesNotMatch",
                "Message": "Computer Over. Virus = Very Yes.",
            }
        },
    )

    def setup_loader_modules(self):
        return {
            boto_route53: {
                "__utils__": {
                    "boto.get_error": MagicMock(return_value="There was an error"),
                },
            },
        }

    def setUp(self):
        # This would normally be set by __utils__["boto.assign_funcs"], but
        # we're not running that as part of this test class, so we need to make
        # sure this attribute is present so that it can be mocked.
        boto_route53._get_conn = None  # pylint: disable=unmocked-patch

    def tearDown(self):
        delattr(boto_route53, "_get_conn")

    def test_zone_exists(self):
        """
        Tests retry behavior for zone_exists
        """
        # Retryable error (max retries reached)
        conn = DummyConn(get_zone=[self._retryable_error, self._retryable_error])
        with patch.object(boto_route53, "_get_conn", MagicMock(return_value=conn)):
            result = boto_route53.zone_exists(
                "foo",
                error_retries=2,
            )
            assert conn.get_zone.call_count == 2
            assert result is False

        # Retryable error (passes on 2nd attempt)
        conn = DummyConn(get_zone=[self._retryable_error, True])
        with patch.object(boto_route53, "_get_conn", MagicMock(return_value=conn)):
            result = boto_route53.zone_exists("foo")
            assert conn.get_zone.call_count == 2
            assert result is True

        # Non-retryable error (should re-raise DNSServerError)
        conn = DummyConn(get_zone=[self._fatal_error])
        with patch.object(boto_route53, "_get_conn", MagicMock(return_value=conn)):
            try:
                result = boto_route53.zone_exists("foo")
            except DNSServerError:
                # This is the expected result
                pass
            else:
                raise Exception("DNSServerError not raised")

    @patch.object(boto, "route53", MagicMock())
    def test_create_healthcheck(self):
        """
        Tests retry behavior for create_healthcheck
        """
        # Retryable error (max retries reached)
        conn = DummyConn(
            create_health_check=[self._retryable_error, self._retryable_error]
        )
        with patch.object(boto_route53, "_get_conn", MagicMock(return_value=conn)):
            result = boto_route53.create_healthcheck(
                "foo",
                error_retries=2,
            )
            assert conn.create_health_check.call_count == 2
            assert result is False

        # Retryable error (passes on 2nd attempt)
        conn = DummyConn(create_health_check=[self._retryable_error, True])
        with patch.object(boto_route53, "_get_conn", MagicMock(return_value=conn)):
            result = boto_route53.create_healthcheck("foo")
            assert conn.create_health_check.call_count == 2
            assert result == {"result": True}, result

        # Non-retryable error (should re-raise DNSServerError)
        conn = DummyConn(create_health_check=[self._fatal_error])
        with patch.object(boto_route53, "_get_conn", MagicMock(return_value=conn)):
            result = boto_route53.create_healthcheck("foo")
            assert conn.create_health_check.call_count == 1
            assert result == {"error": "There was an error"}, result

    def test_get_record(self):
        """
        Tests retry behavior for get_record
        """
        # Retryable error (max retries reached)
        conn = DummyConn(get_zone=[self._retryable_error, self._retryable_error])
        with patch.object(boto_route53, "_get_conn", MagicMock(return_value=conn)):
            result = boto_route53.get_record(
                "foo",
                "bar",
                "baz",
                error_retries=2,
            )
            assert conn.get_zone.call_count == 2
            assert not result

        # Retryable error (passes on 2nd attempt)
        conn = DummyConn(
            get_zone=[
                self._retryable_error,
                namedtuple("Zone", "find_records")(lambda *args, **kwargs: False),
            ]
        )
        with patch.object(boto_route53, "_get_conn", MagicMock(return_value=conn)):
            result = boto_route53.get_record("foo", "bar", "baz")
            assert conn.get_zone.call_count == 2
            assert not result

        # Non-retryable error (should re-raise DNSServerError)
        conn = DummyConn(get_zone=[self._fatal_error])
        with patch.object(boto_route53, "_get_conn", MagicMock(return_value=conn)):
            try:
                result = boto_route53.get_record("foo", "bar", "baz")
            except DNSServerError:
                # This is the expected result
                pass
            else:
                raise Exception("DNSServerError not raised")

    @patch.object(boto_route53, "_wait_for_sync", MagicMock(return_value=True))
    def test_add_record(self):
        """
        Tests retry behavior for add_record
        """
        # Retryable error (max retries reached)
        zone = DummyConn(add_record=[self._retryable_error, self._retryable_error])
        zone.id = "foo"
        conn = DummyConn(get_zone=[zone])
        with patch.object(boto_route53, "_get_conn", MagicMock(return_value=conn)):
            result = boto_route53.add_record(
                "a",
                "b",
                "c",
                "d",
                error_retries=2,
            )
            assert zone.add_record.call_count == 2
            assert not result

        # Retryable error (passes on 2nd attempt)
        zone = DummyConn(
            add_record=[self._retryable_error, namedtuple("Status", "id")("foo")]
        )
        conn = DummyConn(get_zone=[zone])
        with patch.object(boto_route53, "_get_conn", MagicMock(return_value=conn)):
            result = boto_route53.add_record("a", "b", "c", "d")
            assert zone.add_record.call_count == 2
            assert result

        # Non-retryable error (should re-raise DNSServerError)
        zone = DummyConn(add_record=[self._fatal_error])
        conn = DummyConn(get_zone=[zone])
        with patch.object(boto_route53, "_get_conn", MagicMock(return_value=conn)):
            try:
                result = boto_route53.add_record("a", "b", "c", "d")
            except DNSServerError:
                # This is the expected result
                pass
            else:
                raise Exception("DNSServerError not raised")

    @patch.object(boto_route53, "_wait_for_sync", MagicMock(return_value=True))
    def test_update_record(self):
        """
        Tests retry behavior for update_record
        """
        # Retryable error (max retries reached)
        zone = DummyConn(find_records=[self._retryable_error, self._retryable_error])
        conn = DummyConn(get_zone=[zone])
        with patch.object(boto_route53, "_get_conn", MagicMock(return_value=conn)):
            result = boto_route53.update_record(
                "a",
                "b",
                "c",
                "d",
                error_retries=2,
            )
            assert zone.find_records.call_count == 2
            assert not result

        # Retryable error (passes on 2nd attempt)
        zone = DummyConn(
            find_records=[True, True],
            update_record=[self._retryable_error, namedtuple("Status", "id")("foo")],
        )
        conn = DummyConn(get_zone=[zone])
        with patch.object(boto_route53, "_get_conn", MagicMock(return_value=conn)):
            result = boto_route53.update_record("a", "b", "c", "d")
            assert zone.update_record.call_count == 2
            assert result

        # Non-retryable error (should re-raise DNSServerError)
        zone = DummyConn(find_records=[self._fatal_error])
        conn = DummyConn(get_zone=[zone])
        with patch.object(boto_route53, "_get_conn", MagicMock(return_value=conn)):
            try:
                result = boto_route53.update_record("a", "b", "c", "d")
            except DNSServerError:
                # This is the expected result
                pass
            else:
                raise Exception("DNSServerError not raised")

    @patch.object(boto_route53, "_wait_for_sync", MagicMock(return_value=True))
    def test_delete_record(self):
        """
        Tests retry behavior for delete_record
        """
        # Retryable error (max retries reached)
        zone = DummyConn(find_records=[self._retryable_error, self._retryable_error])
        conn = DummyConn(get_zone=[zone])
        with patch.object(boto_route53, "_get_conn", MagicMock(return_value=conn)):
            result = boto_route53.delete_record(
                "a",
                "b",
                "c",
                "d",
                error_retries=2,
            )
            assert zone.find_records.call_count == 2
            assert not result

        # Retryable error (passes on 2nd attempt)
        zone = DummyConn(
            find_records=[True, True],
            delete_record=[self._retryable_error, namedtuple("Status", "id")("foo")],
        )
        conn = DummyConn(get_zone=[zone])
        with patch.object(boto_route53, "_get_conn", MagicMock(return_value=conn)):
            result = boto_route53.delete_record("a", "b", "c", "d")
            assert zone.delete_record.call_count == 2
            assert result

        # Non-retryable error (should re-raise DNSServerError)
        zone = DummyConn(find_records=[self._fatal_error])
        conn = DummyConn(get_zone=[zone])
        with patch.object(boto_route53, "_get_conn", MagicMock(return_value=conn)):
            try:
                result = boto_route53.delete_record("a", "b", "c", "d")
            except DNSServerError:
                # This is the expected result
                pass
            else:
                raise Exception("DNSServerError not raised")



import logging

import pytest
from tests.support.mock import MagicMock, patch

try:
    import dateutil.parser

    HAS_DATEUTIL_PARSER = True
except ImportError:
    HAS_DATEUTIL_PARSER = False

try:
    import croniter  # pylint: disable=unused-import

    HAS_CRONITER = True
except ImportError:
    HAS_CRONITER = False

log = logging.getLogger(__name__)

pytestmark = [
    pytest.mark.skipif(
        HAS_DATEUTIL_PARSER is False,
        reason="The 'dateutil.parser' library is not available",
    ),
]


@pytest.mark.skipif(not HAS_CRONITER, reason="Cannot find croniter python module")
def test_eval_cron_invalid(schedule):
    """
    verify that scheduled job runs
    """
    job = {"schedule": {"job1": {"function": "test.ping", "cron": "0 16 29 13 *"}}}

    # Add the job to the scheduler
    schedule.opts.update(job)

    run_time = dateutil.parser.parse("11/29/2017 4:00pm")
    with patch("croniter.croniter.get_next", MagicMock(return_value=run_time)):
        schedule.eval(now=run_time)

    ret = schedule.job_status("job1")
    assert ret["_error"] == "Invalid cron string. Ignoring job job1."


def test_eval_when_invalid_date(schedule):
    """
    verify that scheduled job does not run
    and returns the right error
    """
    run_time = dateutil.parser.parse("11/29/2017 4:00pm")

    job = {"schedule": {"job1": {"function": "test.ping", "when": "13/29/2017 1:00pm"}}}

    # Add the job to the scheduler
    schedule.opts.update(job)

    # Evaluate 1 second before the run time
    schedule.eval(now=run_time)
    ret = schedule.job_status("job1")
    assert ret["_error"] == "Invalid date string 13/29/2017 1:00pm. Ignoring job job1."


def test_eval_whens_grain_not_dict(schedule):
    """
    verify that scheduled job does not run
    and returns the right error
    """
    schedule.opts["grains"]["whens"] = {"tea time": "11/29/2017 12:00pm"}

    run_time = dateutil.parser.parse("11/29/2017 4:00pm")

    job = {"schedule": {"job1": {"function": "test.ping", "when": "tea time"}}}

    schedule.opts["grains"]["whens"] = ["tea time"]

    # Add the job to the scheduler
    schedule.opts.update(job)

    # Evaluate 1 second before the run time
    schedule.eval(now=run_time)
    ret = schedule.job_status("job1")
    assert ret["_error"] == 'Grain "whens" must be a dict. Ignoring job job1.'


def test_eval_once_invalid_datestring(schedule):
    """
    verify that scheduled job does not run
    and returns the right error
    """
    job = {
        "schedule": {"job1": {"function": "test.ping", "once": "2017-13-13T13:00:00"}}
    }
    run_time = dateutil.parser.parse("12/13/2017 1:00pm")

    # Add the job to the scheduler
    schedule.opts.update(job)

    # Evaluate 1 second at the run time
    schedule.eval(now=run_time)
    ret = schedule.job_status("job1")
    _expected = (
        "Date string could not be parsed: "
        "2017-13-13T13:00:00, %Y-%m-%dT%H:%M:%S. "
        "Ignoring job job1."
    )
    assert ret["_error"] == _expected


def test_eval_skip_during_range_invalid_date(schedule):
    """
    verify that scheduled job does not run
    and returns the right error
    """

    job = {
        "schedule": {
            "job1": {
                "function": "test.ping",
                "hours": 1,
                "skip_during_range": {"start": "1:00pm", "end": "25:00pm"},
            }
        }
    }

    # Add the job to the scheduler
    schedule.opts.update(job)

    # eval at 3:00pm to prime, simulate minion start up.
    run_time = dateutil.parser.parse("11/29/2017 3:00pm")
    schedule.eval(now=run_time)
    ret = schedule.job_status("job1")

    # eval at 4:00pm to prime
    run_time = dateutil.parser.parse("11/29/2017 4:00pm")
    schedule.eval(now=run_time)
    ret = schedule.job_status("job1")
    _expected = "Invalid date string for end in skip_during_range. Ignoring job job1."
    assert ret["_error"] == _expected


def test_eval_skip_during_range_end_before_start(schedule):
    """
    verify that scheduled job does not run
    and returns the right error
    """

    job = {
        "schedule": {
            "job1": {
                "function": "test.ping",
                "hours": 1,
                "skip_during_range": {"start": "1:00pm", "end": "12:00pm"},
            }
        }
    }

    # Add the job to the scheduler
    schedule.opts.update(job)

    # eval at 3:00pm to prime, simulate minion start up.
    run_time = dateutil.parser.parse("11/29/2017 3:00pm")
    schedule.eval(now=run_time)
    ret = schedule.job_status("job1")

    # eval at 4:00pm to prime
    run_time = dateutil.parser.parse("11/29/2017 4:00pm")
    schedule.eval(now=run_time)
    ret = schedule.job_status("job1")
    _expected = (
        "schedule.handle_func: Invalid "
        "range, end must be larger than "
        "start. Ignoring job job1."
    )
    assert ret["_error"] == _expected


def test_eval_skip_during_range_not_dict(schedule):
    """
    verify that scheduled job does not run
    and returns the right error
    """

    job = {
        "schedule": {
            "job1": {
                "function": "test.ping",
                "hours": 1,
                "skip_during_range": ["start", "1:00pm", "end", "12:00pm"],
            }
        }
    }

    # Add the job to the scheduler
    schedule.opts.update(job)

    # eval at 3:00pm to prime, simulate minion start up.
    run_time = dateutil.parser.parse("11/29/2017 3:00pm")
    schedule.eval(now=run_time)
    ret = schedule.job_status("job1")

    # eval at 4:00pm to prime
    run_time = dateutil.parser.parse("11/29/2017 4:00pm")
    schedule.eval(now=run_time)
    ret = schedule.job_status("job1")
    _expected = (
        "schedule.handle_func: Invalid, "
        "range must be specified as a "
        "dictionary. Ignoring job job1."
    )
    assert ret["_error"] == _expected



"""
    :copyright: Copyright 2013-2017 by the SaltStack Team, see AUTHORS for more details.
    :license: Apache 2.0, see LICENSE for more details.


    tests.support.helpers
    ~~~~~~~~~~~~~~~~~~~~~

    Test support helpers
"""
import base64
import builtins
import errno
import fnmatch
import functools
import inspect
import json
import logging
import os
import pathlib
import random
import shutil
import socket
import string
import subprocess
import sys
import tempfile
import textwrap
import threading
import time
import types
from contextlib import contextmanager

import attr
import pytest
import salt.ext.tornado.ioloop
import salt.ext.tornado.web
import salt.utils.files
import salt.utils.platform
import salt.utils.pycrypto
import salt.utils.stringutils
import salt.utils.versions
from saltfactories.exceptions import FactoryFailure as ProcessFailed
from saltfactories.utils.ports import get_unused_localhost_port
from saltfactories.utils.processes import ProcessResult
from tests.support.mock import patch
from tests.support.runtests import RUNTIME_VARS
from tests.support.sminion import create_sminion
from tests.support.unit import SkipTest, _id, skip

log = logging.getLogger(__name__)

HAS_SYMLINKS = None


PRE_PYTEST_SKIP_OR_NOT = "PRE_PYTEST_DONT_SKIP" not in os.environ
PRE_PYTEST_SKIP_REASON = (
    "PRE PYTEST - This test was skipped before running under pytest"
)
PRE_PYTEST_SKIP = pytest.mark.skipif(
    PRE_PYTEST_SKIP_OR_NOT, reason=PRE_PYTEST_SKIP_REASON
)
ON_PY35 = sys.version_info < (3, 6)


def no_symlinks():
    """
    Check if git is installed and has symlinks enabled in the configuration.
    """
    global HAS_SYMLINKS
    if HAS_SYMLINKS is not None:
        return not HAS_SYMLINKS
    output = ""
    try:
        output = subprocess.Popen(
            ["git", "config", "--get", "core.symlinks"],
            cwd=RUNTIME_VARS.TMP,
            stdout=subprocess.PIPE,
        ).communicate()[0]
    except OSError as exc:
        if exc.errno != errno.ENOENT:
            raise
    except subprocess.CalledProcessError:
        # git returned non-zero status
        pass
    HAS_SYMLINKS = False
    if output.strip() == "true":
        HAS_SYMLINKS = True
    return not HAS_SYMLINKS


def destructiveTest(caller):
    """
    Mark a test case as a destructive test for example adding or removing users
    from your system.

    .. code-block:: python

        class MyTestCase(TestCase):

            @destructiveTest
            def test_create_user(self):
                pass
    """
    salt.utils.versions.warn_until_date(
        "20220101",
        "Please stop using `@destructiveTest`, it will be removed in {date}, and"
        " instead use `@pytest.mark.destructive_test`.",
        stacklevel=3,
    )
    setattr(caller, "__destructive_test__", True)

    if os.environ.get("DESTRUCTIVE_TESTS", "False").lower() == "false":
        reason = "Destructive tests are disabled"

        if not isinstance(caller, type):

            @functools.wraps(caller)
            def skip_wrapper(*args, **kwargs):
                raise SkipTest(reason)

            caller = skip_wrapper

        caller.__unittest_skip__ = True
        caller.__unittest_skip_why__ = reason

    return caller


def expensiveTest(caller):
    """
    Mark a test case as an expensive test, for example, a test which can cost
    money(Salt's cloud provider tests).

    .. code-block:: python

        class MyTestCase(TestCase):

            @expensiveTest
            def test_create_user(self):
                pass
    """
    salt.utils.versions.warn_until_date(
        "20220101",
        "Please stop using `@expensiveTest`, it will be removed in {date}, and instead"
        " use `@pytest.mark.expensive_test`.",
        stacklevel=3,
    )
    setattr(caller, "__expensive_test__", True)

    if os.environ.get("EXPENSIVE_TESTS", "False").lower() == "false":
        reason = "Expensive tests are disabled"

        if not isinstance(caller, type):

            @functools.wraps(caller)
            def skip_wrapper(*args, **kwargs):
                raise SkipTest(reason)

            caller = skip_wrapper

        caller.__unittest_skip__ = True
        caller.__unittest_skip_why__ = reason

    return caller


def slowTest(caller):
    """
    Mark a test case as a slow test.
    .. code-block:: python
        class MyTestCase(TestCase):
            @slowTest
            def test_that_takes_much_time(self):
                pass
    """
    salt.utils.versions.warn_until_date(
        "20220101",
        "Please stop using `@slowTest`, it will be removed in {date}, and instead use "
        "`@pytest.mark.slow_test`.",
        stacklevel=3,
    )
    setattr(caller, "__slow_test__", True)
    return caller


def flaky(caller=None, condition=True, attempts=4):
    """
    Mark a test as flaky. The test will attempt to run five times,
    looking for a successful run. After an immediate second try,
    it will use an exponential backoff starting with one second.

    .. code-block:: python

        class MyTestCase(TestCase):

        @flaky
        def test_sometimes_works(self):
            pass
    """
    salt.utils.versions.warn_until_date(
        "20220101",
        "Please stop using `@flaky`, it will be removed in {date}, and instead use "
        "`@pytest.mark.flaky`. See https://pypi.org/project/flaky for information on "
        "how to use it.",
        stacklevel=3,
    )
    if caller is None:
        return functools.partial(flaky, condition=condition, attempts=attempts)

    if isinstance(condition, bool) and condition is False:
        # Don't even decorate
        return caller
    elif callable(condition):
        if condition() is False:
            # Don't even decorate
            return caller

    if inspect.isclass(caller):
        attrs = [n for n in dir(caller) if n.startswith("test_")]
        for attrname in attrs:
            try:
                function = getattr(caller, attrname)
                if not inspect.isfunction(function) and not inspect.ismethod(function):
                    continue
                setattr(
                    caller,
                    attrname,
                    flaky(caller=function, condition=condition, attempts=attempts),
                )
            except Exception as exc:  # pylint: disable=broad-except
                log.exception(exc)
                continue
        return caller

    @functools.wraps(caller)
    def wrap(cls):
        for attempt in range(0, attempts):
            try:
                if attempt > 0:
                    # Run through setUp again
                    # We only run it after the first iteration(>0) because the regular
                    # test runner will have already ran setUp the first time
                    setup = getattr(cls, "setUp", None)
                    if callable(setup):
                        setup()
                return caller(cls)
            except SkipTest as exc:
                cls.skipTest(exc.args[0])
            except Exception as exc:  # pylint: disable=broad-except
                exc_info = sys.exc_info()
                if isinstance(exc, SkipTest):
                    raise exc_info[0].with_traceback(exc_info[1], exc_info[2])
                if not isinstance(exc, AssertionError) and log.isEnabledFor(
                    logging.DEBUG
                ):
                    log.exception(exc, exc_info=exc_info)
                if attempt >= attempts - 1:
                    # We won't try to run tearDown once the attempts are exhausted
                    # because the regular test runner will do that for us
                    raise exc_info[0].with_traceback(exc_info[1], exc_info[2])
                # Run through tearDown again
                teardown = getattr(cls, "tearDown", None)
                if callable(teardown):
                    teardown()
                backoff_time = attempt ** 2
                log.info("Found Exception. Waiting %s seconds to retry.", backoff_time)
                time.sleep(backoff_time)
        return cls

    return wrap


def requires_sshd_server(caller):
    """
    Mark a test as requiring the tests SSH daemon running.

    .. code-block:: python

        class MyTestCase(TestCase):

            @requiresSshdServer
            def test_create_user(self):
                pass
    """
    raise RuntimeError(
        "Please replace @requires_sshd_server with @pytest.mark.requires_sshd_server"
    )


class RedirectStdStreams:
    """
    Temporarily redirect system output to file like objects.
    Default is to redirect to `os.devnull`, which just mutes output, `stdout`
    and `stderr`.
    """

    def __init__(self, stdout=None, stderr=None):
        if stdout is None:
            # pylint: disable=resource-leakage
            stdout = salt.utils.files.fopen(os.devnull, "w")
            # pylint: enable=resource-leakage
        if stderr is None:
            # pylint: disable=resource-leakage
            stderr = salt.utils.files.fopen(os.devnull, "w")
            # pylint: enable=resource-leakage

        self.__stdout = stdout
        self.__stderr = stderr
        self.__redirected = False
        self.patcher = patch.multiple(sys, stderr=self.__stderr, stdout=self.__stdout)

    def __enter__(self):
        self.redirect()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.unredirect()

    def redirect(self):
        self.old_stdout = sys.stdout
        self.old_stdout.flush()
        self.old_stderr = sys.stderr
        self.old_stderr.flush()
        self.patcher.start()
        self.__redirected = True

    def unredirect(self):
        if not self.__redirected:
            return
        try:
            self.__stdout.flush()
            self.__stdout.close()
        except ValueError:
            # already closed?
            pass
        try:
            self.__stderr.flush()
            self.__stderr.close()
        except ValueError:
            # already closed?
            pass
        self.patcher.stop()

    def flush(self):
        if self.__redirected:
            try:
                self.__stdout.flush()
            except Exception:  # pylint: disable=broad-except
                pass
            try:
                self.__stderr.flush()
            except Exception:  # pylint: disable=broad-except
                pass


class TstSuiteLoggingHandler:
    """
    Simple logging handler which can be used to test if certain logging
    messages get emitted or not:

    .. code-block:: python

        with TstSuiteLoggingHandler() as handler:
            # (...)               Do what ever you wish here
            handler.messages    # here are the emitted log messages

    """

    def __init__(self, level=0, format="%(levelname)s:%(message)s"):
        self.level = level
        self.format = format
        self.activated = False
        self.prev_logging_level = None

    def activate(self):
        class Handler(logging.Handler):
            def __init__(self, level):
                logging.Handler.__init__(self, level)
                self.messages = []

            def emit(self, record):
                self.messages.append(self.format(record))

        self.handler = Handler(self.level)
        formatter = logging.Formatter(self.format)
        self.handler.setFormatter(formatter)
        logging.root.addHandler(self.handler)
        self.activated = True
        # Make sure we're running with the lowest logging level with our
        # tests logging handler
        current_logging_level = logging.root.getEffectiveLevel()
        if current_logging_level > logging.DEBUG:
            self.prev_logging_level = current_logging_level
            logging.root.setLevel(0)

    def deactivate(self):
        if not self.activated:
            return
        logging.root.removeHandler(self.handler)
        # Restore previous logging level if changed
        if self.prev_logging_level is not None:
            logging.root.setLevel(self.prev_logging_level)

    @property
    def messages(self):
        if not self.activated:
            return []
        return self.handler.messages

    def clear(self):
        self.handler.messages = []

    def __enter__(self):
        self.activate()
        return self

    def __exit__(self, type, value, traceback):
        self.deactivate()
        self.activated = False

    # Mimic some handler attributes and methods
    @property
    def lock(self):
        if self.activated:
            return self.handler.lock

    def createLock(self):
        if self.activated:
            return self.handler.createLock()

    def acquire(self):
        if self.activated:
            return self.handler.acquire()

    def release(self):
        if self.activated:
            return self.handler.release()


class ForceImportErrorOn:
    """
    This class is meant to be used in mock'ed test cases which require an
    ``ImportError`` to be raised.

    >>> import os.path
    >>> with ForceImportErrorOn('os.path'):
    ...     import os.path
    ...
    Traceback (most recent call last):
      File "<stdin>", line 2, in <module>
      File "salttesting/helpers.py", line 263, in __import__
        'Forced ImportError raised for {0!r}'.format(name)
    ImportError: Forced ImportError raised for 'os.path'
    >>>


    >>> with ForceImportErrorOn(('os', 'path')):
    ...     import os.path
    ...     sys.modules.pop('os', None)
    ...     from os import path
    ...
    <module 'os' from '/usr/lib/python2.7/os.pyc'>
    Traceback (most recent call last):
      File "<stdin>", line 4, in <module>
      File "salttesting/helpers.py", line 288, in __fake_import__
        name, ', '.join(fromlist)
    ImportError: Forced ImportError raised for 'from os import path'
    >>>


    >>> with ForceImportErrorOn(('os', 'path'), 'os.path'):
    ...     import os.path
    ...     sys.modules.pop('os', None)
    ...     from os import path
    ...
    Traceback (most recent call last):
      File "<stdin>", line 2, in <module>
      File "salttesting/helpers.py", line 281, in __fake_import__
        'Forced ImportError raised for {0!r}'.format(name)
    ImportError: Forced ImportError raised for 'os.path'
    >>>
    """

    def __init__(self, *module_names):
        self.__module_names = {}
        for entry in module_names:
            if isinstance(entry, (list, tuple)):
                modname = entry[0]
                self.__module_names[modname] = set(entry[1:])
            else:
                self.__module_names[entry] = None
        self.__original_import = builtins.__import__
        self.patcher = patch.object(builtins, "__import__", self.__fake_import__)

    def patch_import_function(self):
        self.patcher.start()

    def restore_import_funtion(self):
        self.patcher.stop()

    def __fake_import__(
        self, name, globals_=None, locals_=None, fromlist=None, level=None
    ):
        if level is None:
            level = 0
        if fromlist is None:
            fromlist = []

        if name in self.__module_names:
            importerror_fromlist = self.__module_names.get(name)
            if importerror_fromlist is None:
                raise ImportError("Forced ImportError raised for {!r}".format(name))

            if importerror_fromlist.intersection(set(fromlist)):
                raise ImportError(
                    "Forced ImportError raised for {!r}".format(
                        "from {} import {}".format(name, ", ".join(fromlist))
                    )
                )
        return self.__original_import(name, globals_, locals_, fromlist, level)

    def __enter__(self):
        self.patch_import_function()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.restore_import_funtion()


class MockWraps:
    """
    Helper class to be used with the mock library.
    To be used in the ``wraps`` keyword of ``Mock`` or ``MagicMock`` where you
    want to trigger a side effect for X times, and afterwards, call the
    original and un-mocked method.

    As an example:

    >>> def original():
    ...     print 'original'
    ...
    >>> def side_effect():
    ...     print 'side effect'
    ...
    >>> mw = MockWraps(original, 2, side_effect)
    >>> mw()
    side effect
    >>> mw()
    side effect
    >>> mw()
    original
    >>>

    """

    def __init__(self, original, expected_failures, side_effect):
        self.__original = original
        self.__expected_failures = expected_failures
        self.__side_effect = side_effect
        self.__call_counter = 0

    def __call__(self, *args, **kwargs):
        try:
            if self.__call_counter < self.__expected_failures:
                if isinstance(self.__side_effect, types.FunctionType):
                    return self.__side_effect()
                raise self.__side_effect
            return self.__original(*args, **kwargs)
        finally:
            self.__call_counter += 1


def requires_network(only_local_network=False):
    """
    Simple decorator which is supposed to skip a test case in case there's no
    network connection to the internet.
    """
    salt.utils.versions.warn_until_date(
        "20220101",
        "Please stop using `@requires_network`, it will be removed in {date}, and"
        " instead use `@pytest.mark.requires_network`.",
        stacklevel=3,
    )

    def decorator(func):
        @functools.wraps(func)
        def wrapper(cls, *args, **kwargs):
            has_local_network = False
            # First lets try if we have a local network. Inspired in
            # verify_socket
            try:
                pubsock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                retsock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                pubsock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
                pubsock.bind(("", 18000))
                pubsock.close()
                retsock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
                retsock.bind(("", 18001))
                retsock.close()
                has_local_network = True
            except OSError:
                # I wonder if we just have IPV6 support?
                try:
                    pubsock = socket.socket(socket.AF_INET6, socket.SOCK_STREAM)
                    retsock = socket.socket(socket.AF_INET6, socket.SOCK_STREAM)
                    pubsock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
                    pubsock.bind(("", 18000))
                    pubsock.close()
                    retsock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
                    retsock.bind(("", 18001))
                    retsock.close()
                    has_local_network = True
                except OSError:
                    # Let's continue
                    pass

            if only_local_network is True:
                if has_local_network is False:
                    # Since we're only supposed to check local network, and no
                    # local network was detected, skip the test
                    cls.skipTest("No local network was detected")
                return func(cls)

            if os.environ.get("NO_INTERNET"):
                cls.skipTest("Environment variable NO_INTERNET is set.")

            # We are using the google.com DNS records as numerical IPs to avoid
            # DNS lookups which could greatly slow down this check
            for addr in (
                "173.194.41.198",
                "173.194.41.199",
                "173.194.41.200",
                "173.194.41.201",
                "173.194.41.206",
                "173.194.41.192",
                "173.194.41.193",
                "173.194.41.194",
                "173.194.41.195",
                "173.194.41.196",
                "173.194.41.197",
            ):
                sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                try:
                    sock.settimeout(0.25)
                    sock.connect((addr, 80))
                    # We connected? Stop the loop
                    break
                except OSError:
                    # Let's check the next IP
                    continue
                else:
                    cls.skipTest("No internet network connection was detected")
                finally:
                    sock.close()
            return func(cls, *args, **kwargs)

        return wrapper

    return decorator


def with_system_user(
    username, on_existing="delete", delete=True, password=None, groups=None
):
    """
    Create and optionally destroy a system user to be used within a test
    case. The system user is created using the ``user`` salt module.

    The decorated testcase function must accept 'username' as an argument.

    :param username: The desired username for the system user.
    :param on_existing: What to do when the desired username is taken. The
      available options are:

      * nothing: Do nothing, act as if the user was created.
      * delete: delete and re-create the existing user
      * skip: skip the test case
    """
    if on_existing not in ("nothing", "delete", "skip"):
        raise RuntimeError(
            "The value of 'on_existing' can only be one of, "
            "'nothing', 'delete' and 'skip'"
        )

    if not isinstance(delete, bool):
        raise RuntimeError("The value of 'delete' can only be 'True' or 'False'")

    def decorator(func):
        @functools.wraps(func)
        def wrap(cls):

            # Let's add the user to the system.
            log.debug("Creating system user %r", username)
            kwargs = {"timeout": 60, "groups": groups}
            if salt.utils.platform.is_windows():
                kwargs.update({"password": password})
            create_user = cls.run_function("user.add", [username], **kwargs)
            if not create_user:
                log.debug("Failed to create system user")
                # The user was not created
                if on_existing == "skip":
                    cls.skipTest("Failed to create system user {!r}".format(username))

                if on_existing == "delete":
                    log.debug("Deleting the system user %r", username)
                    delete_user = cls.run_function(
                        "user.delete", [username, True, True]
                    )
                    if not delete_user:
                        cls.skipTest(
                            "A user named {!r} already existed on the "
                            "system and re-creating it was not possible".format(
                                username
                            )
                        )
                    log.debug("Second time creating system user %r", username)
                    create_user = cls.run_function("user.add", [username], **kwargs)
                    if not create_user:
                        cls.skipTest(
                            "A user named {!r} already existed, was deleted "
                            "as requested, but re-creating it was not possible".format(
                                username
                            )
                        )
            if not salt.utils.platform.is_windows() and password is not None:
                if salt.utils.platform.is_darwin():
                    hashed_password = password
                else:
                    hashed_password = salt.utils.pycrypto.gen_hash(password=password)
                hashed_password = "'{}'".format(hashed_password)
                add_pwd = cls.run_function(
                    "shadow.set_password", [username, hashed_password]
                )

            failure = None
            try:
                try:
                    return func(cls, username)
                except Exception as exc:  # pylint: disable=W0703
                    log.error(
                        "Running %r raised an exception: %s", func, exc, exc_info=True
                    )
                    # Store the original exception details which will be raised
                    # a little further down the code
                    failure = sys.exc_info()
            finally:
                if delete:
                    delete_user = cls.run_function(
                        "user.delete", [username, True, True], timeout=60
                    )
                    if not delete_user:
                        if failure is None:
                            log.warning(
                                "Although the actual test-case did not fail, "
                                "deleting the created system user %r "
                                "afterwards did.",
                                username,
                            )
                        else:
                            log.warning(
                                "The test-case failed and also did the removal"
                                " of the system user %r",
                                username,
                            )
                if failure is not None:
                    # If an exception was thrown, raise it
                    raise failure[1].with_traceback(failure[2])

        return wrap

    return decorator


def with_system_group(group, on_existing="delete", delete=True):
    """
    Create and optionally destroy a system group to be used within a test
    case. The system user is crated using the ``group`` salt module.

    The decorated testcase function must accept 'group' as an argument.

    :param group: The desired group name for the system user.
    :param on_existing: What to do when the desired username is taken. The
      available options are:

      * nothing: Do nothing, act as if the group was created
      * delete: delete and re-create the existing user
      * skip: skip the test case
    """
    if on_existing not in ("nothing", "delete", "skip"):
        raise RuntimeError(
            "The value of 'on_existing' can only be one of, "
            "'nothing', 'delete' and 'skip'"
        )

    if not isinstance(delete, bool):
        raise RuntimeError("The value of 'delete' can only be 'True' or 'False'")

    def decorator(func):
        @functools.wraps(func)
        def wrap(cls):

            # Let's add the user to the system.
            log.debug("Creating system group %r", group)
            create_group = cls.run_function("group.add", [group])
            if not create_group:
                log.debug("Failed to create system group")
                # The group was not created
                if on_existing == "skip":
                    cls.skipTest("Failed to create system group {!r}".format(group))

                if on_existing == "delete":
                    log.debug("Deleting the system group %r", group)
                    delete_group = cls.run_function("group.delete", [group])
                    if not delete_group:
                        cls.skipTest(
                            "A group named {!r} already existed on the "
                            "system and re-creating it was not possible".format(group)
                        )
                    log.debug("Second time creating system group %r", group)
                    create_group = cls.run_function("group.add", [group])
                    if not create_group:
                        cls.skipTest(
                            "A group named {!r} already existed, was deleted "
                            "as requested, but re-creating it was not possible".format(
                                group
                            )
                        )

            failure = None
            try:
                try:
                    return func(cls, group)
                except Exception as exc:  # pylint: disable=W0703
                    log.error(
                        "Running %r raised an exception: %s", func, exc, exc_info=True
                    )
                    # Store the original exception details which will be raised
                    # a little further down the code
                    failure = sys.exc_info()
            finally:
                if delete:
                    delete_group = cls.run_function("group.delete", [group])
                    if not delete_group:
                        if failure is None:
                            log.warning(
                                "Although the actual test-case did not fail, "
                                "deleting the created system group %r "
                                "afterwards did.",
                                group,
                            )
                        else:
                            log.warning(
                                "The test-case failed and also did the removal"
                                " of the system group %r",
                                group,
                            )
                if failure is not None:
                    # If an exception was thrown, raise it
                    raise failure[1].with_traceback(failure[2])

        return wrap

    return decorator


def with_system_user_and_group(username, group, on_existing="delete", delete=True):
    """
    Create and optionally destroy a system user and group to be used within a
    test case. The system user is crated using the ``user`` salt module, and
    the system group is created with the ``group`` salt module.

    The decorated testcase function must accept both the 'username' and 'group'
    arguments.

    :param username: The desired username for the system user.
    :param group: The desired name for the system group.
    :param on_existing: What to do when the desired username is taken. The
      available options are:

      * nothing: Do nothing, act as if the user was created.
      * delete: delete and re-create the existing user
      * skip: skip the test case
    """
    if on_existing not in ("nothing", "delete", "skip"):
        raise RuntimeError(
            "The value of 'on_existing' can only be one of, "
            "'nothing', 'delete' and 'skip'"
        )

    if not isinstance(delete, bool):
        raise RuntimeError("The value of 'delete' can only be 'True' or 'False'")

    def decorator(func):
        @functools.wraps(func)
        def wrap(cls):

            # Let's add the user to the system.
            log.debug("Creating system user %r", username)
            create_user = cls.run_function("user.add", [username])
            log.debug("Creating system group %r", group)
            create_group = cls.run_function("group.add", [group])
            if not create_user:
                log.debug("Failed to create system user")
                # The user was not created
                if on_existing == "skip":
                    cls.skipTest("Failed to create system user {!r}".format(username))

                if on_existing == "delete":
                    log.debug("Deleting the system user %r", username)
                    delete_user = cls.run_function(
                        "user.delete", [username, True, True]
                    )
                    if not delete_user:
                        cls.skipTest(
                            "A user named {!r} already existed on the "
                            "system and re-creating it was not possible".format(
                                username
                            )
                        )
                    log.debug("Second time creating system user %r", username)
                    create_user = cls.run_function("user.add", [username])
                    if not create_user:
                        cls.skipTest(
                            "A user named {!r} already existed, was deleted "
                            "as requested, but re-creating it was not possible".format(
                                username
                            )
                        )
            if not create_group:
                log.debug("Failed to create system group")
                # The group was not created
                if on_existing == "skip":
                    cls.skipTest("Failed to create system group {!r}".format(group))

                if on_existing == "delete":
                    log.debug("Deleting the system group %r", group)
                    delete_group = cls.run_function("group.delete", [group])
                    if not delete_group:
                        cls.skipTest(
                            "A group named {!r} already existed on the "
                            "system and re-creating it was not possible".format(group)
                        )
                    log.debug("Second time creating system group %r", group)
                    create_group = cls.run_function("group.add", [group])
                    if not create_group:
                        cls.skipTest(
                            "A group named {!r} already existed, was deleted "
                            "as requested, but re-creating it was not possible".format(
                                group
                            )
                        )

            failure = None
            try:
                try:
                    return func(cls, username, group)
                except Exception as exc:  # pylint: disable=W0703
                    log.error(
                        "Running %r raised an exception: %s", func, exc, exc_info=True
                    )
                    # Store the original exception details which will be raised
                    # a little further down the code
                    failure = sys.exc_info()
            finally:
                if delete:
                    delete_user = cls.run_function(
                        "user.delete", [username, True, True]
                    )
                    delete_group = cls.run_function("group.delete", [group])
                    if not delete_user:
                        if failure is None:
                            log.warning(
                                "Although the actual test-case did not fail, "
                                "deleting the created system user %r "
                                "afterwards did.",
                                username,
                            )
                        else:
                            log.warning(
                                "The test-case failed and also did the removal"
                                " of the system user %r",
                                username,
                            )
                    if not delete_group:
                        if failure is None:
                            log.warning(
                                "Although the actual test-case did not fail, "
                                "deleting the created system group %r "
                                "afterwards did.",
                                group,
                            )
                        else:
                            log.warning(
                                "The test-case failed and also did the removal"
                                " of the system group %r",
                                group,
                            )
                if failure is not None:
                    # If an exception was thrown, raise it
                    raise failure[1].with_traceback(failure[2])

        return wrap

    return decorator


class WithTempfile:
    def __init__(self, **kwargs):
        self.create = kwargs.pop("create", True)
        if "dir" not in kwargs:
            kwargs["dir"] = RUNTIME_VARS.TMP
        if "prefix" not in kwargs:
            kwargs["prefix"] = "__salt.test."
        self.kwargs = kwargs

    def __call__(self, func):
        self.func = func
        return functools.wraps(func)(
            # pylint: disable=unnecessary-lambda
            lambda testcase, *args, **kwargs: self.wrap(testcase, *args, **kwargs)
            # pylint: enable=unnecessary-lambda
        )

    def wrap(self, testcase, *args, **kwargs):
        name = salt.utils.files.mkstemp(**self.kwargs)
        if not self.create:
            os.remove(name)
        try:
            return self.func(testcase, name, *args, **kwargs)
        finally:
            try:
                os.remove(name)
            except OSError:
                pass


with_tempfile = WithTempfile


class WithTempdir:
    def __init__(self, **kwargs):
        self.create = kwargs.pop("create", True)
        if "dir" not in kwargs:
            kwargs["dir"] = RUNTIME_VARS.TMP
        self.kwargs = kwargs

    def __call__(self, func):
        self.func = func
        return functools.wraps(func)(
            # pylint: disable=unnecessary-lambda
            lambda testcase, *args, **kwargs: self.wrap(testcase, *args, **kwargs)
            # pylint: enable=unnecessary-lambda
        )

    def wrap(self, testcase, *args, **kwargs):
        tempdir = tempfile.mkdtemp(**self.kwargs)
        if not self.create:
            os.rmdir(tempdir)
        try:
            return self.func(testcase, tempdir, *args, **kwargs)
        finally:
            shutil.rmtree(tempdir, ignore_errors=True)


with_tempdir = WithTempdir


def requires_system_grains(func):
    """
    Function decorator which loads and passes the system's grains to the test
    case.
    """

    @functools.wraps(func)
    def decorator(*args, **kwargs):
        if not hasattr(requires_system_grains, "__grains__"):
            # Late import
            from tests.support.sminion import build_minion_opts

            opts = build_minion_opts(minion_id="runtests-internal-sminion")
            requires_system_grains.__grains__ = salt.loader.grains(opts)
        kwargs["grains"] = requires_system_grains.__grains__
        return func(*args, **kwargs)

    return decorator


@requires_system_grains
def runs_on(grains=None, **kwargs):
    """
    Skip the test if grains don't match the values passed into **kwargs
    if a kwarg value is a list then skip if the grains don't match any item in the list
    """
    reason = kwargs.pop("reason", None)
    for kw, value in kwargs.items():
        if isinstance(value, list):
            if not any(str(grains.get(kw)).lower() != str(v).lower() for v in value):
                if reason is None:
                    reason = "This test does not run on {}={}".format(
                        kw, grains.get(kw)
                    )
                return skip(reason)
        else:
            if str(grains.get(kw)).lower() != str(value).lower():
                if reason is None:
                    reason = "This test runs on {}={}, not {}".format(
                        kw, value, grains.get(kw)
                    )
                return skip(reason)
    return _id


@requires_system_grains
def not_runs_on(grains=None, **kwargs):
    """
    Reverse of `runs_on`.
    Skip the test if any grains match the values passed into **kwargs
    if a kwarg value is a list then skip if the grains match any item in the list
    """
    reason = kwargs.pop("reason", None)
    for kw, value in kwargs.items():
        if isinstance(value, list):
            if any(str(grains.get(kw)).lower() == str(v).lower() for v in value):
                if reason is None:
                    reason = "This test does not run on {}={}".format(
                        kw, grains.get(kw)
                    )
                return skip(reason)
        else:
            if str(grains.get(kw)).lower() == str(value).lower():
                if reason is None:
                    reason = "This test does not run on {}={}, got {}".format(
                        kw, value, grains.get(kw)
                    )
                return skip(reason)
    return _id


def _check_required_sminion_attributes(sminion_attr, *required_items):
    """
    :param sminion_attr: The name of the sminion attribute to check, such as 'functions' or 'states'
    :param required_items: The items that must be part of the designated sminion attribute for the decorated test
    :return The packages that are not available
    """
    # Late import
    from tests.support.sminion import create_sminion

    required_salt_items = set(required_items)
    sminion = create_sminion(minion_id="runtests-internal-sminion")
    available_items = list(getattr(sminion, sminion_attr))
    not_available_items = set()

    name = "__not_available_{items}s__".format(items=sminion_attr)
    if not hasattr(sminion, name):
        setattr(sminion, name, set())

    cached_not_available_items = getattr(sminion, name)

    for not_available_item in cached_not_available_items:
        if not_available_item in required_salt_items:
            not_available_items.add(not_available_item)
            required_salt_items.remove(not_available_item)

    for required_item_name in required_salt_items:
        search_name = required_item_name
        if "." not in search_name:
            search_name += ".*"
        if not fnmatch.filter(available_items, search_name):
            not_available_items.add(required_item_name)
            cached_not_available_items.add(required_item_name)

    return not_available_items


def requires_salt_states(*names):
    """
    Makes sure the passed salt state is available. Skips the test if not

    .. versionadded:: 3000
    """
    salt.utils.versions.warn_until_date(
        "20220101",
        "Please stop using `@requires_salt_states`, it will be removed in {date}, and"
        " instead use `@pytest.mark.requires_salt_states`.",
        stacklevel=3,
    )
    not_available = _check_required_sminion_attributes("states", *names)
    if not_available:
        return skip("Unavailable salt states: {}".format(*not_available))
    return _id


def requires_salt_modules(*names):
    """
    Makes sure the passed salt module is available. Skips the test if not

    .. versionadded:: 0.5.2
    """
    salt.utils.versions.warn_until_date(
        "20220101",
        "Please stop using `@requires_salt_modules`, it will be removed in {date}, and"
        " instead use `@pytest.mark.requires_salt_modules`.",
        stacklevel=3,
    )
    not_available = _check_required_sminion_attributes("functions", *names)
    if not_available:
        return skip("Unavailable salt modules: {}".format(*not_available))
    return _id


def skip_if_binaries_missing(*binaries, **kwargs):
    salt.utils.versions.warn_until_date(
        "20220101",
        "Please stop using `@skip_if_binaries_missing`, it will be removed in {date},"
        " and instead use `@pytest.mark.skip_if_binaries_missing`.",
        stacklevel=3,
    )
    import salt.utils.path

    if len(binaries) == 1:
        if isinstance(binaries[0], (list, tuple, set, frozenset)):
            binaries = binaries[0]
    check_all = kwargs.pop("check_all", False)
    message = kwargs.pop("message", None)
    if kwargs:
        raise RuntimeError(
            "The only supported keyword argument is 'check_all' and "
            "'message'. Invalid keyword arguments: {}".format(", ".join(kwargs.keys()))
        )
    if check_all:
        for binary in binaries:
            if salt.utils.path.which(binary) is None:
                return skip(
                    "{}The {!r} binary was not found".format(
                        message and "{}. ".format(message) or "", binary
                    )
                )
    elif salt.utils.path.which_bin(binaries) is None:
        return skip(
            "{}None of the following binaries was found: {}".format(
                message and "{}. ".format(message) or "", ", ".join(binaries)
            )
        )
    return _id


def skip_if_not_root(func):
    salt.utils.versions.warn_until_date(
        "20220101",
        "Please stop using `@skip_if_not_root`, it will be removed in {date}, and"
        " instead use `@pytest.mark.skip_if_not_root`.",
        stacklevel=3,
    )
    setattr(func, "__skip_if_not_root__", True)

    if not sys.platform.startswith("win"):
        if os.getuid() != 0:
            func.__unittest_skip__ = True
            func.__unittest_skip_why__ = (
                "You must be logged in as root to run this test"
            )
    else:
        current_user = salt.utils.win_functions.get_current_user()
        if current_user != "SYSTEM":
            if not salt.utils.win_functions.is_admin(current_user):
                func.__unittest_skip__ = True
                func.__unittest_skip_why__ = (
                    "You must be logged in as an Administrator to run this test"
                )
    return func


def repeat(caller=None, condition=True, times=5):
    """
    Repeat a test X amount of times until the first failure.

    .. code-block:: python

        class MyTestCase(TestCase):

        @repeat
        def test_sometimes_works(self):
            pass
    """
    if caller is None:
        return functools.partial(repeat, condition=condition, times=times)

    if isinstance(condition, bool) and condition is False:
        # Don't even decorate
        return caller
    elif callable(condition):
        if condition() is False:
            # Don't even decorate
            return caller

    if inspect.isclass(caller):
        attrs = [n for n in dir(caller) if n.startswith("test_")]
        for attrname in attrs:
            try:
                function = getattr(caller, attrname)
                if not inspect.isfunction(function) and not inspect.ismethod(function):
                    continue
                setattr(
                    caller,
                    attrname,
                    repeat(caller=function, condition=condition, times=times),
                )
            except Exception as exc:  # pylint: disable=broad-except
                log.exception(exc)
                continue
        return caller

    @functools.wraps(caller)
    def wrap(cls):
        for attempt in range(1, times + 1):
            log.info("%s test run %d of %s times", cls, attempt, times)
            caller(cls)
        return cls

    return wrap


def http_basic_auth(login_cb=lambda username, password: False):
    """
    A crude decorator to force a handler to request HTTP Basic Authentication

    Example usage:

    .. code-block:: python

        @http_basic_auth(lambda u, p: u == 'foo' and p == 'bar')
        class AuthenticatedHandler(salt.ext.tornado.web.RequestHandler):
            pass
    """

    def wrapper(handler_class):
        def wrap_execute(handler_execute):
            def check_auth(handler, kwargs):

                auth = handler.request.headers.get("Authorization")

                if auth is None or not auth.startswith("Basic "):
                    # No username/password entered yet, we need to return a 401
                    # and set the WWW-Authenticate header to request login.
                    handler.set_status(401)
                    handler.set_header("WWW-Authenticate", "Basic realm=Restricted")

                else:
                    # Strip the 'Basic ' from the beginning of the auth header
                    # leaving the base64-encoded secret
                    username, password = base64.b64decode(auth[6:]).split(":", 1)

                    if login_cb(username, password):
                        # Authentication successful
                        return
                    else:
                        # Authentication failed
                        handler.set_status(403)

                handler._transforms = []
                handler.finish()

            def _execute(self, transforms, *args, **kwargs):
                check_auth(self, kwargs)
                return handler_execute(self, transforms, *args, **kwargs)

            return _execute

        handler_class._execute = wrap_execute(handler_class._execute)
        return handler_class

    return wrapper


def generate_random_name(prefix, size=6):
    """
    Generates a random name by combining the provided prefix with a randomly generated
    ascii string.

    .. versionadded:: 2018.3.0

    prefix
        The string to prefix onto the randomly generated ascii string.

    size
        The number of characters to generate. Default: 6.
    """
    salt.utils.versions.warn_until_date(
        "20220101",
        "Please replace your call 'generate_random_name({0})' with 'random_string({0},"
        " lowercase=False)' as 'generate_random_name' will be removed after {{date}}".format(
            prefix
        ),
        stacklevel=3,
    )
    return random_string(prefix, size=size, lowercase=False)


def random_string(prefix, size=6, uppercase=True, lowercase=True, digits=True):
    """
    Generates a random string.

    ..versionadded: 3001

    Args:
        prefix(str): The prefix for the random string
        size(int): The size of the random string
        uppercase(bool): If true, include uppercased ascii chars in choice sample
        lowercase(bool): If true, include lowercased ascii chars in choice sample
        digits(bool): If true, include digits in choice sample
    Returns:
        str: The random string
    """
    if not any([uppercase, lowercase, digits]):
        raise RuntimeError(
            "At least one of 'uppercase', 'lowercase' or 'digits' needs to be true"
        )
    choices = []
    if uppercase:
        choices.extend(string.ascii_uppercase)
    if lowercase:
        choices.extend(string.ascii_lowercase)
    if digits:
        choices.extend(string.digits)

    return prefix + "".join(random.choice(choices) for _ in range(size))


class Webserver:
    """
    Starts a tornado webserver on 127.0.0.1 on a random available port

    USAGE:

    .. code-block:: python

        from tests.support.helpers import Webserver

        webserver = Webserver('/path/to/web/root')
        webserver.start()
        webserver.stop()
    """

    def __init__(self, root=None, port=None, wait=5, handler=None, ssl_opts=None):
        """
        root
            Root directory of webserver. If not passed, it will default to the
            location of the base environment of the integration suite's file
            roots (tests/integration/files/file/base/)

        port
            Port on which to listen. If not passed, a random one will be chosen
            at the time the start() function is invoked.

        wait : 5
            Number of seconds to wait for the socket to be open before raising
            an exception

        handler
            Can be used to use a subclass of tornado.web.StaticFileHandler,
            such as when enforcing authentication with the http_basic_auth
            decorator.
        """
        if port is not None and not isinstance(port, int):
            raise ValueError("port must be an integer")

        if root is None:
            root = RUNTIME_VARS.BASE_FILES
        try:
            self.root = os.path.realpath(root)
        except AttributeError:
            raise ValueError("root must be a string")

        self.port = port
        self.wait = wait
        self.handler = (
            handler if handler is not None else salt.ext.tornado.web.StaticFileHandler
        )
        self.web_root = None
        self.ssl_opts = ssl_opts

    def target(self):
        """
        Threading target which stands up the tornado application
        """
        self.ioloop = salt.ext.tornado.ioloop.IOLoop()
        self.ioloop.make_current()
        if self.handler == salt.ext.tornado.web.StaticFileHandler:
            self.application = salt.ext.tornado.web.Application(
                [(r"/(.*)", self.handler, {"path": self.root})]
            )
        else:
            self.application = salt.ext.tornado.web.Application(
                [(r"/(.*)", self.handler)]
            )
        self.application.listen(self.port, ssl_options=self.ssl_opts)
        self.ioloop.start()

    @property
    def listening(self):
        if self.port is None:
            return False
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        return sock.connect_ex(("127.0.0.1", self.port)) == 0

    def url(self, path):
        """
        Convenience function which, given a file path, will return a URL that
        points to that path. If the path is relative, it will just be appended
        to self.web_root.
        """
        if self.web_root is None:
            raise RuntimeError("Webserver instance has not been started")
        err_msg = (
            "invalid path, must be either a relative path or a path within {}".format(
                self.root
            )
        )
        try:
            relpath = (
                path if not os.path.isabs(path) else os.path.relpath(path, self.root)
            )
            if relpath.startswith(".." + os.sep):
                raise ValueError(err_msg)
            return "/".join((self.web_root, relpath))
        except AttributeError:
            raise ValueError(err_msg)

    def start(self):
        """
        Starts the webserver
        """
        if self.port is None:
            self.port = get_unused_localhost_port()

        self.web_root = "http{}://127.0.0.1:{}".format(
            "s" if self.ssl_opts else "", self.port
        )

        self.server_thread = threading.Thread(target=self.target)
        self.server_thread.daemon = True
        self.server_thread.start()

        for idx in range(self.wait + 1):
            if self.listening:
                break
            if idx != self.wait:
                time.sleep(1)
        else:
            raise Exception(
                "Failed to start tornado webserver on 127.0.0.1:{} within "
                "{} seconds".format(self.port, self.wait)
            )

    def stop(self):
        """
        Stops the webserver
        """
        self.ioloop.add_callback(self.ioloop.stop)
        self.server_thread.join()


class SaveRequestsPostHandler(salt.ext.tornado.web.RequestHandler):
    """
    Save all requests sent to the server.
    """

    received_requests = []

    def post(self, *args):  # pylint: disable=arguments-differ
        """
        Handle the post
        """
        self.received_requests.append(self.request)

    def data_received(self):  # pylint: disable=arguments-differ
        """
        Streaming not used for testing
        """
        raise NotImplementedError()


class MirrorPostHandler(salt.ext.tornado.web.RequestHandler):
    """
    Mirror a POST body back to the client
    """

    def post(self, *args):  # pylint: disable=arguments-differ
        """
        Handle the post
        """
        body = self.request.body
        log.debug("Incoming body: %s  Incoming args: %s", body, args)
        self.write(body)

    def data_received(self):  # pylint: disable=arguments-differ
        """
        Streaming not used for testing
        """
        raise NotImplementedError()


def dedent(text, linesep=os.linesep):
    """
    A wrapper around textwrap.dedent that also sets line endings.
    """
    linesep = salt.utils.stringutils.to_unicode(linesep)
    unicode_text = textwrap.dedent(salt.utils.stringutils.to_unicode(text))
    clean_text = linesep.join(unicode_text.splitlines())
    if unicode_text.endswith("\n"):
        clean_text += linesep
    if not isinstance(text, str):
        return salt.utils.stringutils.to_bytes(clean_text)
    return clean_text


class PatchedEnviron:
    def __init__(self, **kwargs):
        self.cleanup_keys = kwargs.pop("__cleanup__", ())
        self.kwargs = kwargs
        self.original_environ = None

    def __enter__(self):
        self.original_environ = os.environ.copy()
        for key in self.cleanup_keys:
            os.environ.pop(key, None)

        # Make sure there are no unicode characters in the self.kwargs if we're
        # on Python 2. These are being added to `os.environ` and causing
        # problems
        if sys.version_info < (3,):
            kwargs = self.kwargs.copy()
            clean_kwargs = {}
            for k in self.kwargs:
                key = k
                if isinstance(key, str):
                    key = key.encode("utf-8")
                if isinstance(self.kwargs[k], str):
                    kwargs[k] = kwargs[k].encode("utf-8")
                clean_kwargs[key] = kwargs[k]
            self.kwargs = clean_kwargs

        os.environ.update(**self.kwargs)
        return self

    def __exit__(self, *args):
        os.environ.clear()
        os.environ.update(self.original_environ)


patched_environ = PatchedEnviron


def _cast_to_pathlib_path(value):
    if isinstance(value, pathlib.Path):
        return value
    return pathlib.Path(str(value))


@attr.s(frozen=True, slots=True)
class VirtualEnv:
    venv_dir = attr.ib(converter=_cast_to_pathlib_path)
    env = attr.ib(default=None)
    system_site_packages = attr.ib(default=False)
    pip_requirement = attr.ib(default="pip>=20.2.4,<21.2", repr=False)
    setuptools_requirement = attr.ib(
        default="setuptools!=50.*,!=51.*,!=52.*", repr=False
    )
    environ = attr.ib(init=False, repr=False)
    venv_python = attr.ib(init=False, repr=False)
    venv_bin_dir = attr.ib(init=False, repr=False)

    @venv_dir.default
    def _default_venv_dir(self):
        return pathlib.Path(tempfile.mkdtemp(dir=RUNTIME_VARS.TMP))

    @environ.default
    def _default_environ(self):
        environ = os.environ.copy()
        if self.env:
            environ.update(self.env)
        return environ

    @venv_python.default
    def _default_venv_python(self):
        # Once we drop Py3.5 we can stop casting to string
        if salt.utils.platform.is_windows():
            return str(self.venv_dir / "Scripts" / "python.exe")
        return str(self.venv_dir / "bin" / "python")

    @venv_bin_dir.default
    def _default_venv_bin_dir(self):
        return pathlib.Path(self.venv_python).parent

    def __enter__(self):
        try:
            self._create_virtualenv()
        except subprocess.CalledProcessError:
            raise AssertionError("Failed to create virtualenv")
        return self

    def __exit__(self, *args):
        shutil.rmtree(str(self.venv_dir), ignore_errors=True)

    def install(self, *args, **kwargs):
        return self.run(self.venv_python, "-m", "pip", "install", *args, **kwargs)

    def uninstall(self, *args, **kwargs):
        return self.run(
            self.venv_python, "-m", "pip", "uninstall", "-y", *args, **kwargs
        )

    def run(self, *args, **kwargs):
        check = kwargs.pop("check", True)
        kwargs.setdefault("cwd", str(self.venv_dir))
        kwargs.setdefault("stdout", subprocess.PIPE)
        kwargs.setdefault("stderr", subprocess.PIPE)
        kwargs.setdefault("universal_newlines", True)
        kwargs.setdefault("env", self.environ)
        proc = subprocess.run(args, check=False, **kwargs)
        ret = ProcessResult(
            exitcode=proc.returncode,
            stdout=proc.stdout,
            stderr=proc.stderr,
            cmdline=proc.args,
        )
        log.debug(ret)
        if check is True:
            try:
                proc.check_returncode()
            except subprocess.CalledProcessError:
                raise ProcessFailed(
                    "Command failed return code check",
                    cmdline=proc.args,
                    stdout=proc.stdout,
                    stderr=proc.stderr,
                    exitcode=proc.returncode,
                )
        return ret

    @staticmethod
    def get_real_python():
        """
        The reason why the virtualenv creation is proxied by this function is mostly
        because under windows, we can't seem to properly create a virtualenv off of
        another virtualenv(we can on linux) and also because, we really don't want to
        test virtualenv creation off of another virtualenv, we want a virtualenv created
        from the original python.
        Also, on windows, we must also point to the virtualenv binary outside the existing
        virtualenv because it will fail otherwise
        """
        try:
            if salt.utils.platform.is_windows():
                return os.path.join(sys.real_prefix, os.path.basename(sys.executable))
            else:
                python_binary_names = [
                    "python{}.{}".format(*sys.version_info),
                    "python{}".format(*sys.version_info),
                    "python",
                ]
                for binary_name in python_binary_names:
                    python = os.path.join(sys.real_prefix, "bin", binary_name)
                    if os.path.exists(python):
                        break
                else:
                    raise AssertionError(
                        "Couldn't find a python binary name under '{}' matching: {}".format(
                            os.path.join(sys.real_prefix, "bin"), python_binary_names
                        )
                    )
                return python
        except AttributeError:
            return sys.executable

    def run_code(self, code_string, **kwargs):
        if code_string.startswith("\n"):
            code_string = code_string[1:]
        code_string = textwrap.dedent(code_string).rstrip()
        log.debug(
            "Code to run passed to python:\n>>>>>>>>>>\n%s\n<<<<<<<<<<", code_string
        )
        return self.run(str(self.venv_python), "-c", code_string, **kwargs)

    def get_installed_packages(self):
        data = {}
        ret = self.run(str(self.venv_python), "-m", "pip", "list", "--format", "json")
        for pkginfo in json.loads(ret.stdout):
            data[pkginfo["name"]] = pkginfo["version"]
        return data

    def _create_virtualenv(self):
        sminion = create_sminion()
        sminion.functions.virtualenv.create(
            str(self.venv_dir),
            python=self.get_real_python(),
            system_site_packages=self.system_site_packages,
        )
        self.install("-U", self.pip_requirement, self.setuptools_requirement)
        log.debug("Created virtualenv in %s", self.venv_dir)


@attr.s(frozen=True, slots=True)
class SaltVirtualEnv(VirtualEnv):
    """
    This is a VirtualEnv implementation which has this salt checkout installed in it
    using static requirements
    """

    def _create_virtualenv(self):
        super()._create_virtualenv()
        self.install(RUNTIME_VARS.CODE_DIR)

    def install(self, *args, **kwargs):
        env = self.environ.copy()
        env.update(kwargs.pop("env", None) or {})
        env["USE_STATIC_REQUIREMENTS"] = "1"
        kwargs["env"] = env
        return super().install(*args, **kwargs)


@contextmanager
def change_cwd(path):
    """
    Context manager helper to change CWD for a with code block and restore
    it at the end
    """
    old_cwd = os.getcwd()
    try:
        os.chdir(path)
        # Do stuff
        yield
    finally:
        # Restore Old CWD
        os.chdir(old_cwd)


@functools.lru_cache(maxsize=1)
def get_virtualenv_binary_path():
    # Under windows we can't seem to properly create a virtualenv off of another
    # virtualenv, we can on linux but we will still point to the virtualenv binary
    # outside the virtualenv running the test suite, if that's the case.
    try:
        real_prefix = sys.real_prefix
        # The above attribute exists, this is a virtualenv
        if salt.utils.platform.is_windows():
            virtualenv_binary = os.path.join(real_prefix, "Scripts", "virtualenv.exe")
        else:
            # We need to remove the virtualenv from PATH or we'll get the virtualenv binary
            # from within the virtualenv, we don't want that
            path = os.environ.get("PATH")
            if path is not None:
                path_items = path.split(os.pathsep)
                for item in path_items[:]:
                    if item.startswith(sys.base_prefix):
                        path_items.remove(item)
                os.environ["PATH"] = os.pathsep.join(path_items)
            virtualenv_binary = salt.utils.path.which("virtualenv")
            if path is not None:
                # Restore previous environ PATH
                os.environ["PATH"] = path
            if not virtualenv_binary.startswith(real_prefix):
                virtualenv_binary = None
        if virtualenv_binary and not os.path.exists(virtualenv_binary):
            # It doesn't exist?!
            virtualenv_binary = None
    except AttributeError:
        # We're not running inside a virtualenv
        virtualenv_binary = None
    return virtualenv_binary
